"""Useful functions for parsing"""

from itertools import chain
from pathlib import Path
from typing import Dict, List
import networkx as nx
from .types import (
    AdjList,
    Alpha,
    Generation,
    GraphName,
    LondonaqGraphName,
    LondonaqLocation,
    LondonaqTimestamp,
    SimpleEdgeFunction,
    SimpleEdgeList,
    Vertex,
)


def build_path_to_londonaq_yaml(londonaq_root: Path, name: LondonaqGraphName) -> Path:
    """Build a filepath to the londonaq yaml file

    Args:
        londonaq_root: Root directory of the londonaq dataset
        name: Londonaq graph name

    Returns:
        Filepath to the londonaq yaml file
    """
    return londonaq_root / name.value / f"{name.value}.yaml"


def build_path_to_londonaq_instance(
    londonaq_root: Path,
    name: LondonaqGraphName,
) -> Path:
    """Build a filepath to a londonaq instance

    Args:
        londonaq_root: Root directory of the londonaq dataset
        name: Londonaq graph name

    Returns:
        Filepath to the londonaq txt
    """
    return londonaq_root / name.value / f"{name.value}.txt"


def build_path_to_oplib_instance(
    oplib_root: Path,
    generation: Generation,
    name: GraphName,
    alpha: int = Alpha.fifty.value,
) -> Path:
    """Build a filepath to a oplib instance

    Args:
        oplib_root: The directory of the clones oplib
        generation: Generation of OPLib instance
        name: Graph instance name
        alpha: Percent of the total cost to set the cost limit to.
            Not useful for instances of Prize-collecting TSPs.
            Default is 50.
            Note if you change to a different value, make sure the file exists

    Returns:
        Path to the OPLib instance
    """
    filename: str = name + "-" + generation.value + "-" + str(alpha) + ".oplib"
    return oplib_root / "instances" / generation.value / filename


def build_path_to_tsplib_instance(tsplib_root: Path, name: GraphName) -> Path:
    """Build a filepath to a tsplib instance

    Args:
        tsplib_root: Directory containing TSP txt instances
        name: Name of the instance

    Returns:
        Filepath to the TSP instance
    """
    filename = name.value + ".tsp"
    return tsplib_root / filename


def edge_attribute_names(G: nx.Graph) -> List[str]:
    """Get the names of all edge attributes

    Args:
        G: Graph

    Returns:
        List of attribute names
    """
    return list(set(chain.from_iterable(d.keys() for *_, d in G.edges(data=True))))


def node_attribute_names(G: nx.Graph) -> List[str]:
    """Get the names of all node attributes

    Args:
        G: Graph

    Returns:
        List of node attribute names
    """
    return list(set(chain.from_iterable(d.keys() for _, d in G.nodes(data=True))))


def londonaq_graph_name(
    location_id: LondonaqLocation, timestamp_id: LondonaqTimestamp
) -> LondonaqGraphName:
    """Get a londonaq graph name"""
    return LondonaqGraphName["laq" + location_id.name + timestamp_id.name]


def londonaq_comment(
    location_id: LondonaqLocation, timestamp_id: LondonaqTimestamp
) -> str:
    """Get a comment for a londonaq dataset"""
    comment = f"A London air quality dataset starting at {location_id.value}. "
    comment += f"The UTC timestamp for the air quality forecast is {timestamp_id.value.isoformat()}"
    return comment


def rename_edge_attributes(
    graph: nx.Graph,
    renaming: Dict[str, str],
    copy_graph: bool = False,
    del_old_attr: bool = False,
) -> nx.Graph:
    """Rename edge attributes

    Args:
        graph: Networkx graph
        renaming: Keys are current attribute names. Values are new attribute names.
        copy_graph: If true, copy the graph before renaming attributes.
        del_old_attr: If true, delete the old edge attribute.

    Returns:
        Graph with renamed attributes. If `copy_graph` is `True`, then the copied graph is returned.
        Otherwise the original graph is returned.
    """
    G = graph.copy() if copy_graph else graph
    for u, v, data in G.edges(data=True):
        for old_name, new_name in renaming.items():
            G.edges[u, v][new_name] = data[old_name]
            if del_old_attr:  # delete the old attribute
                data.pop(old_name)
    return G


def rename_node_attributes(
    graph: nx.Graph,
    renaming: Dict[str, str],
    copy_graph: bool = False,
    del_old_attr: bool = False,
) -> nx.Graph:
    """Rename node attributes

    Args:
        graph: Networkx graph
        renaming: Keys are current attribute names. Values are new attribute names.
        copy_graph: If true, copy the graph before renaming attributes.
        del_old_attr: If true, delete the old node attribute.

    Returns:
        Graph with renamed attributes. If `copy_graph` is `True`, then the copied graph is returned.
        Otherwise the original graph is returned.
    """
    G = graph.copy() if copy_graph else graph
    for u, data in G.nodes(data=True):
        for old_name, new_name in renaming.items():
            G.nodes[u][new_name] = data[old_name]
            if del_old_attr:  # delete the old attribute
                data.pop(old_name)
    return G


def adjacency_list_from_edge_list(edge_list: SimpleEdgeList) -> AdjList:
    """Return a adjacency list from an edge list

    Args:
        edge_list: List of tuples representing edges

    Returns:
        Adjacency list representation
    """
    adj_list: AdjList = {}
    for (u, v) in edge_list:
        if u in adj_list:
            adj_list[u].append(v)
        else:
            adj_list[u] = [v]
    return adj_list


def edge_list_from_adjacency_list(adj_list: Dict[Vertex, List]) -> SimpleEdgeList:
    """Returns an edge list from an adjacency list

    Args:
        adj_list: Adjacency list representation

    Returns:
        Edge list representation
    """
    edge_list: SimpleEdgeList = []
    for u, neighbors in adj_list.items():
        for v in neighbors:
            edge_list.append((u, v))
    return edge_list


AdjWeights = Dict[Vertex, Dict[Vertex, int]]


def adjacency_weights_from_edge_dict(weights: SimpleEdgeFunction) -> AdjWeights:
    """Converts a mapping from edges to weights into an adjacency weight mapping"""
    adj_weights: AdjWeights = {}
    for (u, v), w in weights.items():
        if u not in adj_weights:
            adj_weights[u] = {v: w}
        # elif v not in adj_weights[u]:
        #     adj_weights[u] = {v: w}
        else:
            adj_weights[u][v] = w
    return adj_weights


def edge_dict_from_adjacency_weights(adj_weights: AdjWeights) -> SimpleEdgeFunction:
    """Converts adjacency weight representation to edge function"""
    weights: SimpleEdgeFunction = {}
    for u, neighbor_weights in adj_weights.items():
        for v, w in neighbor_weights.items():
            weights[(u, v)] = w
    return weights



"""Travelling Salesman Problem with Profits library"""

from .converter import (
    asymmetric_from_directed,
    asymmetric_from_undirected,
    biggest_vertex_id_from_graph,
    get_original_from_split_vertex,
    get_original_path_from_split_path,
    head_prize,
    is_split_vertex_pair,
    is_vertex_split_head,
    is_vertex_split_tail,
    split_graph_from_properties,
    split_head,
    split_tail,
    tail_prize,
    to_simple_undirected,
    to_vertex_dataframe,
)
from .complete import is_complete, is_complete_with_self_loops
from .exception import (
    EdgesNotAdjacentException,
    EdgeNotFoundException,
    NotSimpleException,
    NotSimpleCycleException,
    NotSimplePathException,
)
from .metric import metricness, mst_cost, uniform_random_cost
from .problem import BaseTSP, ProfitsProblem, is_pctsp_yes_instance
from .utils import (
    build_path_to_londonaq_instance,
    build_path_to_londonaq_yaml,
    build_path_to_oplib_instance,
    build_path_to_tsplib_instance,
    londonaq_comment,
    londonaq_graph_name,
    rename_edge_attributes,
    rename_node_attributes,
)
from .sparse import sparsify_by_cost, sparsify_uid
from .types import (
    Alpha,
    DisjointPaths,
    Edge,
    EdgeFunction,
    EdgeFunctionName,
    EdgeList,
    EdgeWeightType,
    Generation,
    GraphName,
    LondonaqGraphName,
    LondonaqLocation,
    LondonaqLocationShort,
    LondonaqTimestamp,
    LondonaqTimestampId,
    OptimalSolutionTSP,
    SimpleEdgeFunction,
    SimpleEdgeList,
    Vertex,
    VertexFunction,
    VertexFunctionName,
    VertexList,
    VertexLookup,
)
from .walk import (
    edge_list_from_walk,
    is_simple_cycle,
    is_simple_path,
    is_walk,
    order_edge_list,
    remove_self_loops_from_edge_list,
    reorder_edge_list_from_root,
    total_cost,
    total_cost_networkx,
    total_prize,
    total_prize_of_tour,
    vertex_set_from_edge_list,
    walk_from_edge_list,
)

__all__ = [
    "Alpha",
    "BaseTSP",
    "DisjointPaths",
    "Edge",
    "EdgeFunction",
    "EdgeFunctionName",
    "EdgeList",
    "EdgeWeightType",
    "EdgesNotAdjacentException",
    "EdgeNotFoundException",
    "Generation",
    "GraphName",
    "LondonaqGraphName",
    "LondonaqLocation",
    "LondonaqLocationShort",
    "LondonaqTimestamp",
    "LondonaqTimestampId",
    "NotSimpleException",
    "NotSimpleCycleException",
    "NotSimplePathException",
    "OptimalSolutionTSP",
    "ProfitsProblem",
    "SimpleEdgeFunction",
    "SimpleEdgeList",
    "Vertex",
    "VertexFunction",
    "VertexFunctionName",
    "VertexList",
    "VertexLookup",
    "asymmetric_from_directed",
    "asymmetric_from_undirected",
    "biggest_vertex_id_from_graph",
    "build_path_to_londonaq_instance",
    "build_path_to_londonaq_yaml",
    "build_path_to_oplib_instance",
    "build_path_to_tsplib_instance",
    "complete",
    "converter",
    "edge_list_from_walk",
    "get_original_from_split_vertex",
    "get_original_path_from_split_path",
    "head_prize",
    "is_complete",
    "is_complete_with_self_loops",
    "is_pctsp_yes_instance",
    "is_simple_cycle",
    "is_simple_path",
    "is_split_vertex_pair",
    "is_vertex_split_head",
    "is_vertex_split_tail",
    "is_walk",
    "londonaq_comment",
    "londonaq_graph_name",
    "order_edge_list",
    "metricness",
    "mst_cost",
    "problem",
    "remove_self_loops_from_edge_list",
    "rename_edge_attributes",
    "rename_node_attributes",
    "reorder_edge_list_from_root",
    "sparsify_by_cost",
    "sparsify_uid",
    "split_graph_from_properties",
    "split_head",
    "split_tail",
    "tail_prize",
    "to_simple_undirected",
    "to_vertex_dataframe",
    "total_cost",
    "total_cost_networkx",
    "total_prize",
    "total_prize_of_tour",
    "types",
    "uniform_random_cost",
    "utils",
    "walk",
    "walk_from_edge_list",
    "vertex_set_from_edge_list",
]



"""Fixtures for testing"""

import networkx as nx
import pytest
from tspwplib import EdgeFunctionName, VertexFunctionName, VertexList

# pylint: disable=redefined-outer-name


@pytest.fixture(scope="function", params=[[], [0], [0, 1, 3, 1, 2]])
def open_walk(request) -> VertexList:
    """Open walk that may repeat vertices"""
    return request.param


@pytest.fixture(scope="function", params=[[], [0], [0, 1, 3, 1, 0]])
def closed_walk(request) -> VertexList:
    """Walk that starts and ends at same vertex"""
    return request.param


@pytest.fixture(scope="function", params=[[0, 2, 1], [0, 4, 1]])
def not_a_walk(request) -> VertexList:
    """Not a walk"""
    return request.param


@pytest.fixture(scope="function", params=[[], [0], [0, 1, 3, 0]])
def simple_cycle(request) -> VertexList:
    """Simple cycle"""
    return request.param


@pytest.fixture(scope="function", params=[[], [0], [0, 1, 2]])
def simple_path(request) -> VertexList:
    """Simple path"""
    return request.param


@pytest.fixture(scope="function")
def walk_graph() -> nx.Graph:
    """Graph for testing walks"""
    return nx.Graph([(0, 1), (0, 3), (1, 2), (1, 3)])


@pytest.fixture(scope="function")
def weighted_walk_networkx_graph(walk_graph) -> nx.Graph:
    """Walk graph with cost on edges and prize on vertices"""
    for vertex in walk_graph:
        walk_graph.nodes[vertex][VertexFunctionName.prize] = vertex
    edge_attrs = {}
    for u, v in walk_graph.edges():
        edge_attrs[(u, v)] = {EdgeFunctionName.cost.value: u + v}
    nx.set_edge_attributes(walk_graph, edge_attrs)
    return walk_graph



"""Test that all existing instances are complete graphs"""

import networkx as nx
import tsplib95
from tspwplib import (
    is_complete,
    is_complete_with_self_loops,
    build_path_to_tsplib_instance,
)


def test_is_complete():
    """Test the is complete function"""
    for k in range(0, 10):
        graph = nx.complete_graph(k)
        assert is_complete(graph)
        assert sum(range(graph.number_of_nodes())) == graph.number_of_edges()


def test_is_complete_with_self_loops():
    """Test graph is complete and has self loops"""
    for k in range(1, 10):
        graph = nx.complete_graph(k)
        assert is_complete(graph)
        assert not is_complete_with_self_loops(graph)
        for u in graph:
            graph.add_edge(u, u)
        assert is_complete_with_self_loops(graph)


def test_tsplib_is_complete(tsplib_root, graph_name):
    """Test each instance of TSPLIB95 is a complete graph"""
    filepath = build_path_to_tsplib_instance(tsplib_root, graph_name)
    problem = tsplib95.load(filepath)
    graph = problem.get_graph()
    assert is_complete_with_self_loops(graph)



"""Tests for creating metric and non-metric cost functions"""

import networkx as nx
import pytest
from tspwplib import build_path_to_oplib_instance, metricness, mst_cost, ProfitsProblem
from tspwplib.metric import semi_mst_cost


def test_mst_cost(oplib_root, generation, graph_name):
    """Test MST cost"""
    filepath = build_path_to_oplib_instance(oplib_root, generation, graph_name)
    problem = ProfitsProblem.load(filepath)
    G = problem.get_graph()
    new_cost = mst_cost(G, cost_attr="cost")
    T = nx.minimum_spanning_tree(G, weight="cost")

    for (u, v), cost in nx.get_edge_attributes(G, "cost").items():
        if u == v:
            assert cost == 0
        elif T.has_edge(u, v):
            assert new_cost[(u, v)] == cost
        else:
            assert new_cost[(u, v)] == cost + nx.shortest_path_length(
                T, u, v, weight="cost"
            )


def test_semi_mst_cost(oplib_root, generation, graph_name):
    """Test Semi MST cost"""
    filepath = build_path_to_oplib_instance(oplib_root, generation, graph_name)
    problem = ProfitsProblem.load(filepath)
    G = problem.get_graph()
    new_cost = semi_mst_cost(G, cost_attr="cost")
    T = nx.minimum_spanning_tree(G, weight="cost")
    tree_cost = sum(nx.get_edge_attributes(T, "cost").values())
    upper_bound_cost = sum(mst_cost(G).values())
    assert tree_cost < sum(new_cost.values()) < upper_bound_cost


@pytest.mark.parametrize(
    "edges,expected_metricness",
    [
        ([(0, 1, 1), (1, 2, 1), (2, 0, 5)], 0),
        ([(0, 1, 1), (1, 2, 1), (2, 0, 1)], 1),
        ([(0, 1, 1), (1, 2, 1), (2, 0, 5), (2, 3, 1)], 0.0),
        (
            [
                (0, 1, 1),
                (1, 2, 1),
                (2, 0, 5),
                (2, 3, 1),
                (3, 4, 1),
                (4, 2, 1),
                (1, 3, 5),
            ],
            1.0 / 3.0,
        ),
        (
            [
                (0, 1, 10),
                (1, 2, 1),
                (2, 3, 1),
                (3, 4, 1),
                (1, 4, 1),
                (4, 5, 1),
                (0, 5, 1),
            ],
            0.5,
        ),
    ],
)
def test_metricness(edges, expected_metricness):
    """Test metricness"""
    G = nx.Graph()
    G.add_weighted_edges_from(edges, weight="cost")
    assert metricness(G) == expected_metricness



"""Tests for the utils module"""

import itertools
import networkx as nx
import pytest
from tspwplib import (
    londonaq_graph_name,
    LondonaqGraphName,
    LondonaqLocation,
    LondonaqTimestamp,
    build_path_to_oplib_instance,
    build_path_to_tsplib_instance,
    rename_edge_attributes,
    rename_node_attributes,
)


def test_build_path_to_oplib_instance(oplib_root, generation, graph_name):
    """Test path building"""
    assert build_path_to_oplib_instance(oplib_root, generation, graph_name).exists()


def test_build_path_to_tsplib_instance(tsplib_root, graph_name):
    """Test path building for tsplib"""
    assert build_path_to_tsplib_instance(tsplib_root, graph_name).exists()


def test_rename_edge_attributes():
    """Test edge attributes are renamed"""
    G = nx.from_edgelist([(0, 1), (1, 2)])
    nx.set_edge_attributes(G, 1, "weight")
    rename_edge_attributes(G, {"weight": "cost"})
    for _, _, data in G.edges(data=True):
        assert "weight" in data
        assert "cost" in data
        assert data["weight"] == data["cost"]

    # now copy the graph and delete the old edge attribute
    new_graph = rename_edge_attributes(
        G, {"weight": "length", "cost": "height"}, copy_graph=True, del_old_attr=True
    )
    for _, _, data in new_graph.edges(data=True):
        assert not "weight" in data
        assert not "cost" in data
        assert "length" in data
        assert "height" in data
        assert data["length"] == data["height"]
    for _, _, data in G.edges(data=True):
        assert "weight" in data
        assert "cost" in data
        assert not "length" in data
        assert not "height" in data


def test_rename_node_attributes():
    """Test node attributes are renamed"""
    G = nx.from_edgelist([(0, 1), (1, 2)])
    nx.set_node_attributes(G, 1, "weight")
    rename_node_attributes(G, {"weight": "cost"})
    for _, data in G.nodes(data=True):
        assert "weight" in data
        assert "cost" in data
        assert data["weight"] == data["cost"]

    # now copy the graph and delete the old node attribute
    new_graph = rename_node_attributes(
        G, {"weight": "length", "cost": "height"}, copy_graph=True, del_old_attr=True
    )
    for _, data in new_graph.nodes(data=True):
        assert not "weight" in data
        assert not "cost" in data
        assert "length" in data
        assert "height" in data
        assert data["length"] == data["height"]
    for _, data in G.nodes(data=True):
        assert "weight" in data
        assert "cost" in data
        assert not "length" in data
        assert not "height" in data


@pytest.mark.parametrize(
    "location,timestamp", itertools.product(LondonaqLocation, LondonaqTimestamp)
)
def test_londonaq_graph_name(location, timestamp):
    """Test every combination of location and timestamp yields a valid graph name"""
    assert londonaq_graph_name(location, timestamp) in LondonaqGraphName



"""Type hinting and names"""

from datetime import datetime, timezone
from enum import Enum, IntEnum
from typing import Any, Dict, List, Tuple, Union

# vertex data structures
Vertex = int
VertexFunction = Dict[Vertex, int]
VertexList = List[Vertex]
VertexLookup = Dict[Vertex, Vertex]
VertexProperties = Dict[Vertex, Dict[str, Any]]

# edge data structures
Edge = Tuple[Vertex, Vertex]
MultiEdge = Tuple[Vertex, Vertex, int]
EdgeList = List[Union[Edge, MultiEdge]]
EdgeFunction = Dict[Union[Edge, MultiEdge], int]
EdgeProperties = Dict[Union[Edge, MultiEdge], Dict[str, Any]]
SimpleEdgeFunction = Dict[Edge, int]
SimpleEdgeList = List[Edge]

# path data structures
DisjointPaths = Tuple[VertexList, VertexList]

# graph data structures
AdjList = Dict[Vertex, VertexList]

# pylint: disable=invalid-name,too-few-public-methods


class StrEnumMixin:
    """When the `str(...)` method is called on this mixin, return the value of the Enum."""

    def __str__(self):
        try:
            return self.value()
        except TypeError:
            return self


class EdgeWeightType(StrEnumMixin, str, Enum):
    """Specifies how the edge weights (or distances) are given"""

    EXPLICIT = "EXPLICIT"  # Weights are listed explicitly in the corresponding section
    EUC_2D = "EUC_2D"  # Weights are Euclidean distances in 2-D
    EUC_3D = "EUC_3D"  # Weights are Euclidean distances in 3-D
    MAX_2D = "MAX_2D"  # Weights are maximum distances in 2-D
    MAX_3D = "MAX_3D"  # Weights are maximum distances in 3-D
    MAN_2D = "MAN_2D"  # Weights are Manhattan distances in 2-D
    MAN_3D = "MAN_3D"  # Weights are Manhattan distances in 3-D
    CEIL_2D = "CEIL_2D"  # Weights are Euclidean distances in 2-D rounded up
    GEO = "GEO"  # Weights are geographical distances
    ATT = "ATT"  # Special distance function for problems att48 and att532
    XRAY1 = (
        "XRAY1"  # Special distance function for crystallography problems (Version 1)
    )
    XRAY2 = (
        "XRAY2"  # Special distance function for crystallography problems (Version 2)
    )
    SPECIAL = "SPECIAL"  # There is a special distance function documented elsewhere
    UNIFORM_RANDOM = "UNIFORM_RANDOM"
    MST = "MST"


class EdgeWeightFormat(StrEnumMixin, str, Enum):
    """Describes the format of the edge weights if they are given explicitly"""

    FUNCTION = "FUNCTION"  # Weights are given by a function (see above)
    FULL_MATRIX = "FULL_MATRIX"  # Weights are given by a full matrix
    UPPER_ROW = (
        "UPPER_ROW"  # Upper triangular matrix (row-wise without diagonal entries)
    )
    LOWER_ROW = (
        "LOWER_ROW"  # Lower triangular matrix (row-wise without diagonal entries)
    )
    UPPER_DIAG_ROW = "UPPER_DIAG_ROW"  # Upper triangular matrix
    LOWER_DIAG_ROW = "LOWER_DIAG_ROW"  # Lower triangular matrix
    UPPER_COL = (
        "UPPER_COL"  # Upper triangular matrix (column-wise without diagonal entries)
    )
    LOWER_COL = (
        "LOWER_COL"  # Lower triangular matrix (column-wise without diagonal entries)
    )
    UPPER_DIAG_COL = "UPPER_DIAG_COL"  # Upper triangular matrix
    LOWER_DIAG_COL = "LOWER_DIAG_COL"  # Lower triangular matrix


class EdgeDataFormat(StrEnumMixin, str, Enum):
    """How the edges are listed.

    Notes:
        This does not include edge attributes. It is only the edge IDs.
    """

    EDGE_LIST = "EDGE_LIST"
    ADJ_LIST = "ADJ_LIST"


class NodeCoordType(StrEnumMixin, str, Enum):
    """How node co-ordinates are represented"""

    TWOD_COORDS = "TWOD_COORDS"  # Nodes are specified by coordinates in 2-D
    THREED_COORDS = "THREED_COORDS"  # Nodes are specified by coordinates in 3-D
    NO_COORDS = "NO_COORDS"  # The nodes do not have associated coordinates


class DisplayDataType(StrEnumMixin, str, Enum):
    """How visualisation should be done."""

    COORD_DISPLAY = "COORD_DISPLAY"  # Display is generated from the node coordinates
    TWOD_DISPLAY = "TWOD_DISPLAY"  # Explicit coordinates in 2-D are given
    NO_DISPLAY = "NO_DISPLAY"  # No graphical display is possible


# NodeCoords = Dict[Vertex, Union[Tuple[float, float], Tuple[float, float, float]]]
NodeCoords = Dict[Vertex, Tuple[float, float]]


class VertexFunctionName(StrEnumMixin, str, Enum):
    """Valid names of functions on vertices"""

    demand = "demand"
    prize = "prize"


class EdgeFunctionName(StrEnumMixin, str, Enum):
    """Valid names of functions on edges"""

    cost = "cost"
    weight = "weight"


class GraphName(StrEnumMixin, str, Enum):
    """Names of TSPlib instances"""

    a280 = "a280"
    ali535 = "ali535"
    att48 = "att48"
    att532 = "att532"
    bayg29 = "bayg29"
    bays29 = "bays29"
    berlin52 = "berlin52"
    bier127 = "bier127"
    brazil58 = "brazil58"
    brd14051 = "brd14051"
    brg180 = "brg180"
    burma14 = "burma14"
    ch130 = "ch130"
    ch150 = "ch150"
    d198 = "d198"
    d493 = "d493"
    d657 = "d657"
    d1291 = "d1291"
    d1655 = "d1655"
    d2103 = "d2103"
    d15112 = "d15112"
    d18512 = "d18512"
    dantzig42 = "dantzig42"
    dsj1000 = "dsj1000"
    eil51 = "eil51"
    eil76 = "eil76"
    eil101 = "eil101"
    fl417 = "fl417"
    fl1400 = "fl1400"
    fl1577 = "fl1577"
    fl3795 = "fl3795"
    fnl4461 = "fnl4461"
    fri26 = "fri26"
    gil262 = "gil262"
    gr17 = "gr17"
    gr21 = "gr21"
    gr24 = "gr24"
    gr48 = "gr48"
    gr96 = "gr96"
    gr120 = "gr120"
    gr137 = "gr137"
    gr202 = "gr202"
    gr229 = "gr229"
    gr431 = "gr431"
    gr666 = "gr666"
    hk48 = "hk48"
    kroA100 = "kroA100"
    kroB100 = "kroB100"
    kroC100 = "kroC100"
    kroD100 = "kroD100"
    kroE100 = "kroE100"
    kroA150 = "kroA150"
    kroB150 = "kroB150"
    kroA200 = "kroA200"
    kroB200 = "kroB200"
    lin105 = "lin105"
    lin318 = "lin318"
    linhp318 = "linhp318"
    nrw1379 = "nrw1379"
    p654 = "p654"
    pa561 = "pa561"
    pcb442 = "pcb442"
    pcb1173 = "pcb1173"
    pcb3038 = "pcb3038"
    pla7397 = "pla7397"
    pla33810 = "pla33810"
    pla85900 = "pla85900"
    pr76 = "pr76"
    pr107 = "pr107"
    pr124 = "pr124"
    pr136 = "pr136"
    pr144 = "pr144"
    pr152 = "pr152"
    pr226 = "pr226"
    pr264 = "pr264"
    pr299 = "pr299"
    pr439 = "pr439"
    pr1002 = "pr1002"
    pr2392 = "pr2392"
    rat99 = "rat99"
    rat195 = "rat195"
    rat575 = "rat575"
    rat783 = "rat783"
    rd100 = "rd100"
    rd400 = "rd400"
    rl1304 = "rl1304"
    rl1323 = "rl1323"
    rl1889 = "rl1889"
    rl5915 = "rl5915"
    rl5934 = "rl5934"
    rl11849 = "rl11849"
    si175 = "si175"
    si535 = "si535"
    si1032 = "si1032"
    st70 = "st70"
    swiss42 = "swiss42"
    ts225 = "ts225"
    tsp225 = "tsp225"
    u159 = "u159"
    u574 = "u574"
    u724 = "u724"
    u1060 = "u1060"
    u1432 = "u1432"
    u1817 = "u1817"
    u2152 = "u2152"
    u2319 = "u2319"
    ulysses16 = "ulysses16"
    ulysses22 = "ulysses22"
    usa13509 = "usa13509"
    vm1084 = "vm1084"
    vm1748 = "vm1748"


class LondonaqGraphName(StrEnumMixin, str, Enum):
    """Names of graphs with London air quality forecasts"""

    laqkxA = "laqkxA"
    laqtinyA = "laqtinyA"
    laqbbA = "laqbbA"
    laqidA = "laqidA"
    laqwsA = "laqwsA"


class LondonaqTimestamp(Enum):
    """Timestamps of the forecasts for London air quality forecasts"""

    A = datetime(2021, 10, 13, 8, 0, 0, tzinfo=timezone.utc)  # 9am BST


class LondonaqTimestampId(StrEnumMixin, str, Enum):
    """Timestamp IDs for CLI"""

    A = "A"
    # B = "B"
    # C = "C"
    # D = "D"


class LondonaqLocation(StrEnumMixin, str, Enum):
    """Names of locations that the London air quality graph is centered upon"""

    bb = "Big Ben"
    kx = "King's Cross"
    tiny = "Camden Town"
    id = "Isle of Dogs"
    ws = "Wembley Stadium"


class LondonaqLocationShort(StrEnumMixin, str, Enum):
    """Short codes for londonaq locations"""

    bb = "bb"
    kx = "kx"
    tiny = "tiny"
    id = "id"
    ws = "ws"


class Generation(StrEnumMixin, str, Enum):
    """Generations of TSPwP problem instances"""

    gen1 = "gen1"
    gen2 = "gen2"
    gen3 = "gen3"


class Alpha(IntEnum):
    """Ratio between profit/cost limit and profit/cost of TSP solution"""

    fifty = 50


class OptimalSolutionTSP(IntEnum):
    """Value of optimal solutions to TSP instances"""

    a280 = 2579
    ali535 = 202339
    att48 = 10628
    att532 = 27686
    bayg29 = 1610
    bays29 = 2020
    berlin52 = 7542
    bier127 = 118282
    brazil58 = 25395
    brd14051 = 469385
    brg180 = 1950
    burma14 = 3323
    ch130 = 6110
    ch150 = 6528
    d198 = 15780
    d493 = 35002
    d657 = 48912
    d1291 = 50801
    d1655 = 62128
    d2103 = 80450
    d15112 = 1573084
    d18512 = 645238
    dantzig42 = 699
    dsj1000 = 18659688  # (EUC_2D)
    # dsj1000 = 18660188 # (CEIL_2D)    # NOTE breaks keys
    eil51 = 426
    eil76 = 538
    eil101 = 629
    fl417 = 11861
    fl1400 = 20127
    fl1577 = 22249
    fl3795 = 28772
    fnl4461 = 182566
    fri26 = 937
    gil262 = 2378
    gr17 = 2085
    gr21 = 2707
    gr24 = 1272
    gr48 = 5046
    gr96 = 55209
    gr120 = 6942
    gr137 = 69853
    gr202 = 40160
    gr229 = 134602
    gr431 = 171414
    gr666 = 294358
    hk48 = 11461
    kroA100 = 21282
    kroB100 = 22141
    kroC100 = 20749
    kroD100 = 21294
    kroE100 = 22068
    kroA150 = 26524
    kroB150 = 26130
    kroA200 = 29368
    kroB200 = 29437
    lin105 = 14379
    lin318 = 42029
    linhp318 = 41345
    nrw1379 = 56638
    p654 = 34643
    pa561 = 2763
    pcb442 = 50778
    pcb1173 = 56892
    pcb3038 = 137694
    pla7397 = 23260728
    pla33810 = 66048945
    pla85900 = 142382641
    pr76 = 108159
    pr107 = 44303
    pr124 = 59030
    pr136 = 96772
    pr144 = 58537
    pr152 = 73682
    pr226 = 80369
    pr264 = 49135
    pr299 = 48191
    pr439 = 107217
    pr1002 = 259045
    pr2392 = 378032
    rat99 = 1211
    rat195 = 2323
    rat575 = 6773
    rat783 = 8806
    rd100 = 7910
    rd400 = 15281
    rl1304 = 252948
    rl1323 = 270199
    rl1889 = 316536
    rl5915 = 565530
    rl5934 = 556045
    rl11849 = 923288
    si175 = 21407
    si535 = 48450
    si1032 = 92650
    st70 = 675
    swiss42 = 1273
    ts225 = 126643
    tsp225 = 3916
    u159 = 42080
    u574 = 36905
    u724 = 41910
    u1060 = 224094
    u1432 = 152970
    u1817 = 57201
    u2152 = 64253
    u2319 = 234256
    ulysses16 = 6859
    ulysses22 = 7013
    usa13509 = 19982859
    vm1084 = 239297
    vm1748 = 336556



"""Test types"""

from tspwplib import GraphName, OptimalSolutionTSP


def test_no_missing_graphs():
    """Iterate through all graph names and check they have a solution"""
    for graph_name in GraphName:
        assert graph_name.name == graph_name.value
        assert graph_name.name in OptimalSolutionTSP.__members__
        assert OptimalSolutionTSP[graph_name.name] > 0



"""Functions for walks in a graph"""

import itertools
from typing import Dict, Iterable, Mapping, Set

import networkx as nx
from .exception import EdgesNotAdjacentException, NotSimpleException
from .types import (
    EdgeFunction,
    EdgeFunctionName,
    EdgeList,
    Vertex,
    VertexList,
    VertexLookup,
)


def edge_list_from_walk(walk: VertexList) -> EdgeList:
    """Get ordered list of edges from an ordered list of vertices

    Args:
        walk: Ordered list of vertices that represent a walk in the graph

    Returns:
        List of edges in the same order as the walk
    """
    edge_list: EdgeList = []
    if len(walk) <= 1:
        return edge_list
    for i in range(len(walk) - 1):
        vertex = walk[i]
        next_vertex = walk[i + 1]
        edge_list.append((vertex, next_vertex))
    return edge_list


def vertex_set_from_edge_list(edge_list: EdgeList) -> Set[Vertex]:
    """Get a set of vertices from a list of edges

    Args:
        edge_list: List of edges

    Returns:
        Set of vertices in the edge list
    """
    return set(itertools.chain.from_iterable(edge_list))


def __add_vertex_to_occurence(
    first_occurence: Dict[Vertex, int],
    second_occurence: Dict[Vertex, int],
    vertex: Vertex,
    index: int,
) -> None:
    """Add vertex to either the first or second occurence lookup"""
    if vertex in first_occurence:
        second_occurence[vertex] = index
    elif vertex not in first_occurence:
        first_occurence[vertex] = index
    else:
        message = f"Vertex {vertex} appears in more than two edges. Walk is not simple."
        raise NotSimpleException(message)


def order_edge_list(unordered_edges: EdgeList) -> EdgeList:
    """Given a list of unordered edges, return an ordered edge list
    such that every two adjacent edge in the list are also adjacent in
    the input graph.

    Note that the list of edges should form a simple path or cycle.

    Args:
        unordered_edges: List of unique edges in no particular order

    Returns:
        List of unique edges that are adjacent in the graph

    Raises:
        NotSimpleException: If the list of edges is not a simple path or cycle
    """
    if not unordered_edges:
        return []
    # create a lookup table of the first and second occurence of each vertex in the edge list
    first_occurence: VertexLookup = {}
    second_occurence: VertexLookup = {}
    for i, edge in enumerate(unordered_edges):
        __add_vertex_to_occurence(first_occurence, second_occurence, edge[0], i)
        __add_vertex_to_occurence(first_occurence, second_occurence, edge[1], i)

    # use the lookup tables to place the edges in the correct order in the edge list
    ordered_edges = []
    j = 0
    target_index = -1
    found_source = False
    first_vertex = 0
    for i, edge in enumerate(unordered_edges):
        u = edge[0]
        v = edge[1]
        if not found_source and u not in second_occurence:
            j = i
            found_source = True
        elif found_source and u not in second_occurence:
            target_index = i
            break
        elif not found_source and v not in second_occurence:
            j = i
            found_source = True
            first_vertex = 1
        elif found_source and v not in second_occurence:
            target_index = i
            break
    prev = unordered_edges[j][first_vertex]
    visited = [False] * len(unordered_edges)

    for i in range(len(unordered_edges)):
        edge = unordered_edges[j]
        if visited[j]:
            raise NotSimpleException()
        visited[j] = True
        u = edge[0]
        v = edge[1]
        ordered_edges.append(edge)
        if j == target_index:
            break

        # if u == prev then follow v
        if u == prev and j == first_occurence[v]:
            j = second_occurence[v]
            prev = v
        elif u == prev and j == second_occurence[v]:
            j = first_occurence[v]
            prev = v
        # if v == prev then follow u
        elif v == prev and j == first_occurence[u]:
            j = second_occurence[u]
            prev = u
        elif v == prev and j == second_occurence[u]:
            j = first_occurence[u]
            prev = u

    return ordered_edges


def reorder_edge_list_from_root(edge_list: EdgeList, root: Vertex) -> EdgeList:
    """Reorder a list of edges such that the root vertex is in the first (and last) edge

    Args:
        edge_list: List of unique, adjacent edges
        root: Root vertex

    Returns:
        List of edges. The first (and last) edge will contain the root vertex.

    Raises:
        nx.NodeNotFound: If the root vertex is not in any edges.
    """
    root_index = -1
    not_found = nx.NodeNotFound(f"Root vertex {root} not found in edge list")
    n = len(edge_list)
    if n == 0:
        raise not_found
    if n > 1 and root in edge_list[0] and root in edge_list[n - 1]:
        return edge_list
    for i in range(n):
        edge = edge_list[i]
        if root in edge:
            root_index = i
    if root_index == -1:
        raise not_found
    reordered_edges = edge_list[root_index:] + edge_list[:root_index]
    return reordered_edges


def walk_from_edge_list(edge_list: EdgeList) -> VertexList:
    """Get a walk from a list of unique, adjacent edges

    Args:
        edge_list: List of unique edges that are adjacent in the graph

    Returns:
        List of vertices in walk of edges

    Raises:
        EdgesNotAdjacentException: When two edges in the walk are not adjacent
    """
    walk: VertexList = []
    if len(edge_list) == 0:
        return walk

    first_edge = edge_list[0]
    if len(edge_list) == 1:
        walk.append(first_edge[0])
        walk.append(first_edge[1])
        return walk

    second_edge = edge_list[1]
    if first_edge[0] == second_edge[0] or first_edge[0] == second_edge[1]:
        current = first_edge[0]
        walk.append(first_edge[1])
    else:
        current = first_edge[1]
        walk.append(first_edge[0])

    for i in range(1, len(edge_list)):
        walk.append(current)
        edge = edge_list[i]
        u = edge[0]
        v = edge[1]
        if u == current:
            current = v
        elif v == current:
            current = u
        else:
            message = f"Edges in the edge list must be adjacent, but edge {u} - {v}"
            message += (
                f" is not adjacent to vertex {current} from previous edge in list."
            )
            raise EdgesNotAdjacentException(message)
    walk.append(current)
    return walk


def is_walk(G: nx.Graph, walk: VertexList) -> bool:
    """Is the walk a sequence of adjacent vertices in the graph?

    Args:
        G: input graph
        walk: Ordered sequence of vertices

    Returns:
        True if all vertices are adjacent in the graph
    """
    edge_list = edge_list_from_walk(walk)
    return all(G.has_edge(edge[0], edge[1]) for edge in edge_list)


def is_simple_cycle(G: nx.Graph, cycle: VertexList) -> bool:
    """Is the cycle simple in the graph?

    Args:
        G: input graph
        cycle: Ordered sequence of vertices

    Returns:
        True if the cycle is simple in the graph
    """
    cycle_length = len(cycle)
    if cycle_length <= 1:
        return True
    return (
        is_walk(G, cycle)
        and cycle_length == len(set(cycle)) + 1
        and cycle[0] == cycle[cycle_length - 1]
    )


def is_simple_path(G: nx.Graph, path: VertexList) -> bool:
    """Is the path simple in the graph?

    Args:
        G: input graph
        path: Ordered sequence of vertices

    Returns:
        True if the path is simple in the graph
    """
    return is_walk(G, path) and len(path) == len(set(path))


def total_prize(prizes: Mapping[Vertex, int], vertices: Iterable[Vertex]) -> int:
    """Total prize of vertices

    Args:
        prizes: A mapping from vertices to prizes, e.g. dict, property map
        vertices: List of vertices in the prizes map

    Returns:
        Total prize of vertices
    """
    sum_prize: int = 0
    for vertex in vertices:
        sum_prize += prizes[vertex]
    return sum_prize


def total_prize_of_tour(prizes: Mapping[Vertex, int], tour: VertexList) -> int:
    """Total prize of unique vertices in the tour

    Args:
        prizes: A mapping from vertices to prizes, e.g. dict, property map
        vertices: List of vertices in the prizes map

    Returns:
        Total prize of the tour
    """
    return total_prize(prizes, set(tour))


def total_cost(costs: EdgeFunction, edges: EdgeList) -> int:
    """Total cost of edges

    Args:
        costs: Mapping from edges to costs
        edges: List of edges

    Returns:
        Total cost of edges
    """
    sum_cost = 0
    for edge in edges:
        try:
            sum_cost += costs[edge]
        except KeyError:
            try:
                u = edge[0]
                v = edge[1]
                sum_cost += costs[(v, u)]
            except KeyError as second_key_error:
                raise KeyError(
                    "Edge ({u},{v}) or ({v},{u}) do not exist in costs map".format(
                        u=u, v=v
                    )
                ) from second_key_error
        except Exception as error:
            raise KeyError(f"{edge} does not exist in cost map") from error
    return sum_cost


def total_cost_networkx(graph: nx.Graph, walk: VertexList) -> int:
    """Get the total cost of edges in a walk of the networkx graph

    Args:
        graph: Undirected input graph with cost attribute
        walk: A sequence of adjacent vertices

    Returns:
        Total cost of edges in the walk
    """
    edges_in_tour = edge_list_from_walk(walk)
    cost_attr = nx.get_edge_attributes(graph, EdgeFunctionName.cost.value)
    return total_cost(cost_attr, edges_in_tour)


def remove_self_loops_from_edge_list(edge_list: EdgeList) -> EdgeList:
    """Return a new edge list with no self loops

    Args:
        edge_list: List of edges

    Returns:
        Edge list with no self loops
    """
    return [edge for edge in edge_list if edge[0] != edge[1]]



"""Function for metric and non-metric cost functions"""

import random
import networkx as nx
import numpy as np
from .exception import NoTreesException, NotConnectedException
from .types import SimpleEdgeList, SimpleEdgeFunction


def metricness(graph: nx.Graph, cost_attr: str = "cost") -> float:
    """Measures how metric a cost function is

    Args:
        graph: Must be undirected, connected and not a tree
        cost_attr: Name of cost attribute

    Returns:
        If a cost function is metric, return 1.0.
        If n-1 edges are metric and the remaining edges are non-metric, return 0.0.

    Raises:
        NotConnectedException: If the graph is not connected
        NoTreesException: If the graph is a tree

    Notes:
        Self loops are ignored from the metricness
    """
    if not nx.is_connected(graph):
        raise NotConnectedException("Make sure your graph is connected")
    if nx.is_tree(graph):
        raise NoTreesException("Make sure your graph is not a tree")
    path_cost = dict(nx.all_pairs_bellman_ford_path_length(graph, weight=cost_attr))
    num_metric = 0
    num_non_metric = 0
    num_self_loops = 0
    for (u, v), cost in nx.get_edge_attributes(graph, cost_attr).items():
        if u == v:
            num_self_loops += 1
        elif cost <= path_cost[u][v]:
            num_metric += 1
        else:
            num_non_metric += 1
    print(num_metric, "metric edges and", num_non_metric, "non metric edges")
    numerator = (float)(num_metric - graph.number_of_nodes() + 1)
    denominator = (float)(
        graph.number_of_edges() - graph.number_of_nodes() - num_self_loops + 1
    )
    return numerator / denominator


def uniform_random_cost(
    edge_list: SimpleEdgeList, min_value: int = 1, max_value: int = 100, seed: int = 0
) -> SimpleEdgeFunction:
    """Generate a cost function for each edge drawn from a uniform and independant probability

    Args:
        edge_list: List of edges in graph
        min_value: Minimum value the cost can take (inclusive)
        max_value: Maximum value the cost can take (inclusive)
        seed: Set the seed of the random number generator

    Returns:
        Edge cost function
    """
    cost: SimpleEdgeFunction = {}
    random.seed(seed)
    for u, v in edge_list:
        cost[(u, v)] = random.randint(min_value, max_value)
    return cost


def mst_cost(G: nx.Graph, cost_attr: str = "cost") -> SimpleEdgeFunction:
    """Find the minimum spanning tree of G.
    The cost of edges in the tree remains unchanged.
    The cost of edges not in the tree is equal to the cost of the minimum spanning tree
    plus the original cost of the edges.

    Args:
        G: Undirected, simple graph
        cost_attr: Name of the cost attribute of edges

    Returns
        A new cost function
    """
    # find the cost of the minimum spanning tree in G
    T = nx.minimum_spanning_tree(G, weight=cost_attr)
    tree_cost = dict(nx.all_pairs_bellman_ford_path_length(T, weight=cost_attr))

    # set the cost of the new edges
    new_cost: SimpleEdgeFunction = {}
    for (u, v), cost in nx.get_edge_attributes(G, cost_attr).items():
        if T.has_edge(u, v):
            new_cost[(u, v)] = cost
        else:
            new_cost[(u, v)] = cost + tree_cost[u][v]
    return new_cost


def semi_mst_cost(
    G: nx.Graph, cost_attr: str = "cost", seed: int = 0
) -> SimpleEdgeFunction:
    """Half of the non-MST-tree edges are left unchanged.
    The other half are assigned cost as described in mst_cost.

    The half of edges are chosen with uniform and independent probability.

    Args:
        G: Undirected, simple graph
        cost_attr: Name of the cost attribute of edges
        seed: Set the seed of the random number generator

    Returns
        A new cost function
    """
    # find the cost of the minimum spanning tree in G
    T = nx.minimum_spanning_tree(G, weight=cost_attr)
    tree_cost = dict(nx.all_pairs_bellman_ford_path_length(T, weight=cost_attr))

    # get edges not in the tree
    non_tree_edges = [(u, v) for u, v in G.edges() if not T.has_edge(u, v)]
    num_non_tree_edges = len(non_tree_edges)

    gen = np.random.default_rng(seed=seed)
    donot_change_edge_cost = set()
    while len(non_tree_edges) >= float(num_non_tree_edges) * 0.5:
        index = gen.integers(0, len(non_tree_edges) - 1)  # get random index
        donot_change_edge_cost.add(
            non_tree_edges.pop(index)
        )  # remove item at random index, add to set

    # set the cost of the new edges
    new_cost: SimpleEdgeFunction = {}
    for (u, v), cost in nx.get_edge_attributes(G, cost_attr).items():
        if (
            T.has_edge(u, v)
            or (u, v) in donot_change_edge_cost
            or (v, u) in donot_change_edge_cost
        ):
            new_cost[(u, v)] = cost
        else:
            new_cost[(u, v)] = cost + tree_cost[u][v]
    return new_cost



"""Functions and classes for datasets"""

from pathlib import Path
from typing import List, Optional, Union, no_type_check
import yaml

import networkx as nx
import numpy as np
import numpy.typing as npt
import pandas as pd
import pydantic
import tsplib95

from .metric import uniform_random_cost
from .tsplib95 import TempEdgeDataField
from .types import (
    DisplayDataType,
    EdgeDataFormat,
    EdgeList,
    EdgeWeightFormat,
    EdgeWeightType,
    NodeCoords,
    NodeCoordType,
    SimpleEdgeFunction,
    SimpleEdgeList,
    Vertex,
    VertexFunction,
    VertexFunctionName,
    VertexList,
    VertexLookup,
)
from .utils import (
    AdjWeights,
    adjacency_list_from_edge_list,
    adjacency_weights_from_edge_dict,
    edge_attribute_names,
    edge_dict_from_adjacency_weights,
    edge_list_from_adjacency_list,
    node_attribute_names,
)
from .walk import is_simple_cycle, walk_from_edge_list, total_prize

# pylint: disable=too-few-public-methods


class BaseTSP(pydantic.BaseModel):
    """A pydantic model for tsplib95.

    Each field is validated with type hinting.
    """

    # pylint: disable=too-many-arguments

    capacity: Optional[Union[int, float]]
    comment: str
    demands: Optional[VertexFunction]
    depots: VertexList
    dimension: int
    display_data: Optional[NodeCoords]
    display_data_type: DisplayDataType
    edge_data: SimpleEdgeList
    edge_data_format: EdgeDataFormat
    edge_weights: Optional[SimpleEdgeFunction]
    edge_weight_format: EdgeWeightFormat
    edge_weight_type: EdgeWeightType
    fixed_edges: SimpleEdgeList
    name: str
    node_coords: Optional[NodeCoords]
    node_coord_type: NodeCoordType
    problem_type: str
    tours: Optional[List[VertexList]]

    class Config:
        """Pydantic configuration"""

        arbitrary_types_allowed = True

    @classmethod
    def from_networkx(
        cls,
        name: str,
        comment: str,
        problem_type: str,
        G: nx.Graph,
        capacity: Optional[Union[int, float]] = None,
        display_data: Optional[NodeCoords] = None,
        display_data_type: DisplayDataType = DisplayDataType.NO_DISPLAY,
        edge_weight_format: EdgeWeightFormat = EdgeWeightFormat.FULL_MATRIX,
        weight_attr_name: str = "weight",
    ):
        """Get a base TSP model from a networkx graph"""
        edge_attr_names = edge_attribute_names(G)
        node_attr_names = node_attribute_names(G)
        if weight_attr_name not in edge_attr_names:
            message = f"{weight_attr_name} is required to be an edge attribute, "
            message += "but was not found in graph. "
            message += "This function only supports an explicit weight function. "
            raise NotImplementedError(message)
        is_2d = "x" in node_attr_names and "y" in node_attr_names
        is_3d = is_2d and "z" in node_attr_names
        if is_3d:
            raise NotImplementedError("3D coords are not supported")
            # node_coord_type = NodeCoordType.THREED_COORDS
            # node_coords = {
            #     node: (float(data["x"]), float(data["y"]), float(data["z"]))
            #     for node, data in G.nodes(data=True)
            # }
        if is_2d:
            node_coord_type = NodeCoordType.TWOD_COORDS
            node_coords = {
                node: (float(data["x"]), float(data["y"]))
                for node, data in G.nodes(data=True)
            }
        else:
            node_coord_type = NodeCoordType.NO_COORDS
            node_coords = {}

        demands = None
        if "demand" in node_attr_names:
            demands = nx.get_node_attributes(G, "demand")
        if display_data_type == DisplayDataType.COORD_DISPLAY:
            display_data = node_coords

        fixed_edges = []
        if "is_fixed" in edge_attr_names:
            fixed_edges = [
                (u, v) for u, v, data in G.edges(data=True) if data["is_fixed"]
            ]

        depots = []
        if "is_depot" in node_attr_names:
            depots = [node for node, data in G.nodes(data=True) if data["is_depot"]]
        edge_data = list(G.edges())
        edge_weights = nx.get_edge_attributes(G, weight_attr_name)
        return cls(
            capacity=capacity,
            comment=comment,
            demands=demands,
            depots=depots,
            dimension=G.number_of_nodes(),
            display_data=display_data,
            display_data_type=display_data_type,
            edge_data=edge_data,
            edge_data_format=EdgeDataFormat.EDGE_LIST,
            edge_weights=edge_weights,
            edge_weight_format=edge_weight_format,
            edge_weight_type=EdgeWeightType.EXPLICIT,
            fixed_edges=fixed_edges,
            name=name,
            node_coords=node_coords,
            node_coord_type=node_coord_type,
            problem_type=problem_type,
            tours=None,
        )

    @classmethod
    def from_dataframes(
        cls,
        name: str,
        comment: str,
        problem_type: str,
        edges_df: pd.DataFrame,
        nodes_df: pd.DataFrame,
        capacity: Optional[Union[int, float]] = None,
        display_data: Optional[NodeCoords] = None,
        display_data_type: DisplayDataType = DisplayDataType.NO_DISPLAY,
        edge_weight_format: EdgeWeightFormat = EdgeWeightFormat.FULL_MATRIX,
    ):
        """Get a TSP base model from edge and node dataframes

        Notes:
            Essential edge columns: [source, target, weight].
            Optional edge columns: [is_fixed].
            Essential node columns: [node, is_depot].
            Optional node columns: [x, y, z, demand].
            The edge weight function is explicitly given by the 'weight' column.
        """
        if "weight" not in edges_df:
            message = "'weight' is not a column in edges_df. "
            message += "This function only supports an explicit weight function. "
            message += "If you have a column that can be used as the weight function, "
            message += "please rename the column to 'weight'."
            raise NotImplementedError(message)
        is_2d = "x" in nodes_df.columns and "y" in nodes_df.columns
        is_3d = is_2d and "z" in nodes_df.columns
        if is_3d:
            raise NotImplementedError("3D coords not supported")
        if is_2d:
            node_coord_type = NodeCoordType.TWOD_COORDS
            node_coords = dict(zip(nodes_df["node"], zip(nodes_df["x"], nodes_df["y"])))
        else:
            node_coord_type = NodeCoordType.NO_COORDS
            node_coords = {}

        demands = None
        if "demand" in nodes_df.columns:
            demands = dict(zip(nodes_df["node"], nodes_df["demand"]))

        if display_data_type == DisplayDataType.COORD_DISPLAY:
            display_data = node_coords

        fixed_edges = []
        if "is_fixed" in edges_df.columns:
            fixed_edges_df = edges_df.loc[edges_df["is_fixed"]]
            fixed_edges = list(zip(fixed_edges_df["source"], fixed_edges_df["target"]))

        depots = nodes_df.loc[nodes_df["is_depot"]]["node"].to_list()
        edge_data = list(zip(edges_df["source"], edges_df["target"]))
        edge_weights = dict(zip(edge_data, edges_df["weight"]))
        return cls(
            capacity=capacity,
            comment=comment,
            demands=demands,
            depots=depots,
            dimension=len(nodes_df["node"]),
            display_data=display_data,
            display_data_type=display_data_type,
            edge_data=edge_data,
            edge_data_format=EdgeDataFormat.EDGE_LIST,
            edge_weights=edge_weights,
            edge_weight_format=edge_weight_format,
            edge_weight_type=EdgeWeightType.EXPLICIT,
            fixed_edges=fixed_edges,
            name=name,
            node_coords=node_coords,
            node_coord_type=node_coord_type,
            problem_type=problem_type,
            tours=None,
        )

    @classmethod
    def from_tsplib95(cls, problem: tsplib95.models.StandardProblem):
        """Get a TSP base model from a StandardProblem object"""

        display_data_type = (
            problem.display_data_type
            if problem.display_data_type
            else DisplayDataType.NO_DISPLAY
        )
        edge_data_format = (
            problem.edge_data_format
            if problem.edge_data_format
            else EdgeDataFormat.EDGE_LIST
        )
        edge_weight_type = problem.edge_weight_type

        # edge weight format
        edge_weight_format = problem.edge_weight_format
        if edge_weight_type == EdgeWeightType.UNIFORM_RANDOM:
            edge_weight_format = EdgeWeightFormat.FUNCTION
            weights = uniform_random_cost(list(problem.get_edges()))
        elif (
            not edge_weight_format
            and edge_weight_type in EdgeWeightType.__members__
            and edge_weight_type != EdgeWeightType.EXPLICIT
        ):
            edge_weight_format = EdgeWeightFormat.FUNCTION
            weights = {(i, j): problem.get_weight(i, j) for i, j in problem.get_edges()}
        elif not edge_weight_format and edge_weight_type == EdgeWeightType.EXPLICIT:
            raise ValueError(
                "Edge weight type is set to EXPLICIT but no edge weight format is given"
            )
        elif not edge_weight_format:
            raise ValueError(
                "Edge weight format in StandardProblem is not set - cannot assign edge weights."
            )
        else:
            weights = {(i, j): problem.get_weight(i, j) for i, j in problem.get_edges()}

        node_coord_type = (
            problem.node_coord_type
            if problem.node_coord_type
            else NodeCoordType.NO_COORDS
        )
        node_coords = None
        if node_coord_type == NodeCoordType.TWOD_COORDS:
            node_coords = {i: problem.node_coords.get(i) for i in problem.get_nodes()}
        elif node_coord_type == NodeCoordType.THREED_COORDS:
            raise NotImplementedError("3D coords not yet supported")

        return cls(
            capacity=problem.capacity,
            comment=problem.comment if problem.comment else "",
            demands=problem.demands,
            depots=problem.depots,
            dimension=problem.dimension,
            display_data=problem.display_data,
            display_data_type=display_data_type,
            edge_data=list(problem.get_edges()),
            edge_data_format=edge_data_format,
            edge_weights=weights,
            edge_weight_format=edge_weight_format,
            edge_weight_type=edge_weight_type,
            fixed_edges=problem.fixed_edges,
            name=problem.name,
            node_coords=node_coords,
            node_coord_type=node_coord_type,
            problem_type=problem.type,
            tours=problem.tours,
        )

    @no_type_check
    def to_tsplib95(self) -> tsplib95.models.StandardProblem:
        """Convert to a tsplib95 standard model"""
        weights = None
        if self.edge_weight_type == EdgeWeightType.EXPLICIT:
            weights = self.get_weighted_full_matrix()

        optional_kwargs = {}
        if not self.capacity is None:
            optional_kwargs["capacity"] = self.capacity
        if self.display_data:
            optional_kwargs["display_data"] = self.display_data
        if self.fixed_edges:
            optional_kwargs["fixed_edges"] = self.fixed_edges
        if self.tours:
            optional_kwargs["tours"] = self.tours
        return tsplib95.models.StandardProblem(
            comment=self.comment,
            demands=self.demands,
            depots=self.depots,
            dimension=self.dimension,
            display_data_type=self.display_data_type,
            edge_data=self.edge_data,
            edge_data_format=self.edge_data_format,
            edge_weights=weights,
            edge_weight_format=self.edge_weight_format,
            edge_weight_type=self.edge_weight_type,
            name=self.name,
            node_coords=self.node_coords,
            node_coord_type=self.node_coord_type,
            type=self.problem_type,
            **optional_kwargs,
        )

    @classmethod
    def from_yaml(cls, yaml_filepath: Path):
        """Load from a yaml file"""
        with open(yaml_filepath, "r", encoding="utf-8") as yaml_file:
            yaml_dict = yaml.load(yaml_file, Loader=yaml.FullLoader)
        edge_data = edge_list_from_adjacency_list(yaml_dict.pop("edge_data"))
        edge_weights = edge_dict_from_adjacency_weights(yaml_dict.pop("edge_weights"))
        fixed_edges = edge_list_from_adjacency_list(yaml_dict.pop("fixed_edges"))
        return cls(
            **yaml_dict,
            edge_data=edge_data,
            edge_weights=edge_weights,
            fixed_edges=fixed_edges,
        )

    def to_yaml(self, yaml_filepath: Path) -> None:
        """Dump the TSP to a YAML file"""
        yaml_dict = dict(
            comment=self.comment,
            demands=self.demands,
            depots=self.depots,
            dimension=self.dimension,
            display_data_type=self.display_data_type.value,
            edge_data=adjacency_list_from_edge_list(self.edge_data),
            edge_data_format=self.edge_data_format.value,
            edge_weight_format=self.edge_weight_format.value,
            edge_weight_type=self.edge_weight_type.value,
            fixed_edges=adjacency_list_from_edge_list(self.fixed_edges),
            name=self.name,
            node_coords=self.node_coords,
            node_coord_type=self.node_coord_type.value,
            problem_type=self.problem_type,
            tours=self.tours,
        )
        if self.edge_weights:
            weights: AdjWeights = adjacency_weights_from_edge_dict(self.edge_weights)
            yaml_dict["edge_weights"] = weights  # type: ignore
        else:
            yaml_dict["edge_weights"] = None
        with open(yaml_filepath, "w", encoding="utf-8") as yaml_file:
            yaml.dump(yaml_dict, yaml_file)

    def get_weighted_full_matrix(self) -> npt.NDArray[np.int_]:
        """Get a square weighted adjacency matrix, sorted by node ID"""
        # create a graph
        G = self.get_graph()
        # sort the node list (ascending)
        sorted_node_list = list(G.nodes())
        sorted_node_list.sort()
        # then get the weighted adjacency matrix
        return nx.to_numpy_array(
            G, nodelist=sorted_node_list, weight="weight", dtype=int
        )

    def __set_graph_attributes(self, graph: nx.Graph) -> None:
        """Set graph attributes such as 'name' and 'comment'"""
        graph.graph["name"] = self.name
        graph.graph["comment"] = self.comment
        graph.graph["problem_type"] = self.problem_type
        graph.graph["dimension"] = self.dimension
        if not self.capacity is None:
            graph.graph["capacity"] = self.capacity

    def __set_node_attributes(self, graph: nx.Graph) -> None:
        """Set node attributes"""
        for vertex in graph.nodes():
            graph.nodes[vertex]["is_depot"] = vertex in self.depots
            if self.demands:
                graph.nodes[vertex]["demand"] = self.demands[vertex]
            if self.display_data:
                graph.nodes[vertex]["display"] = self.display_data[vertex]
            if self.node_coords:
                coords = self.node_coords[vertex]
                graph.nodes[vertex]["x"] = coords[0]
                graph.nodes[vertex]["y"] = coords[1]

    def __add_edges(self, graph: nx.Graph) -> None:
        """Add edges from edge data

        Args:
            graph: Input graph
        """
        for edge in self.edge_data:
            graph.add_edge(edge[0], edge[1])

    def __set_edge_attributes(self, graph: nx.Graph) -> None:
        """Set edge attributes for 'weight' and 'is_fixed'

        Args:
            graph: Input graph
        """
        nx.set_edge_attributes(graph, self.edge_weights, name="weight")
        fixed = {(u, v): (u, v) in self.fixed_edges for u, v in graph.edges()}
        nx.set_edge_attributes(graph, fixed, name="is_fixed")

    def get_graph(self) -> nx.Graph:
        """Get a networkx graph

        Returns:
            Undirected networkx graph with node attributes such as 'is_depot'
            and edge attributes such as 'weight' and 'is_fixed'.
        """
        G = nx.Graph()
        self.__set_graph_attributes(G)
        self.__add_edges(G)
        self.__set_edge_attributes(G)
        self.__set_node_attributes(G)
        return G


class PrizeCollectingTSP(BaseTSP):
    """Prize-collecting TSP pydantic model"""

    def get_root_vertex(self) -> Vertex:
        """Get the root vertex from the 'depots' attribute

        Returns:
            Root vertex

        Raises:
            ValueError: If the number of depots to choose from is zero or greater than 1
        """
        if len(self.depots) > 1:
            raise ValueError(
                "More than 1 depot to choose from: which depot should I choose?"
            )
        try:
            # pylint: disable=unsubscriptable-object
            return self.depots[0]
        except KeyError as key_error:
            raise ValueError("The list of depots is empty") from key_error

    def get_total_prize(self) -> Union[int, float]:
        """ "Get the total prize (demand) of all vertices"""
        if self.demands:
            return sum(self.demands.values())
        return 0


class ProfitsProblem(tsplib95.models.StandardProblem):
    """TSP with Profits Problem"""

    # overwrite edge data to fix bugs
    edge_data = TempEdgeDataField("EDGE_DATA")
    # Maximum distance of the total route in a OP.
    cost_limit = tsplib95.fields.IntegerField("COST_LIMIT")
    # The scores of the nodes of a OP are given in the form (per line)
    node_score = tsplib95.fields.DemandsField("NODE_SCORE_SECTION")
    # The optimal solution to the TSP
    tspsol = tsplib95.fields.IntegerField("TSPSOL")

    def __init__(self, special=None, **data):
        super().__init__(special=special, **data)

    def __set_edge_attributes(self, graph: nx.Graph, names: VertexLookup) -> None:
        """Set edge attributes"""
        # add every edge with some associated metadata
        for edge in self.get_edges():
            cost: int = self.get_weight(edge[0], edge[1])
            # pylint: disable=unsupported-membership-test
            # is_fixed: bool = (u, v) in self.fixed_edges
            graph.add_edge(names[edge[0]], names[edge[1]], cost=cost)

    def __set_graph_attributes(self, graph: nx.Graph) -> None:
        """Set attributes of the graph such as the name"""
        graph.graph["name"] = self.name
        graph.graph["comment"] = self.comment
        graph.graph["type"] = self.type
        graph.graph["dimension"] = self.dimension
        graph.graph["capacity"] = self.capacity
        graph.graph["root"] = self.get_root_vertex()

    def __set_node_attributes(self, graph: nx.Graph, names: VertexLookup) -> None:
        """Add node attributes"""
        node_score = self.get_node_score()
        for vertex in list(self.get_nodes()):
            # pylint: disable=unsupported-membership-test,no-member
            is_depot = vertex in self.depots
            graph.add_node(
                names[vertex],
                prize=node_score[vertex],
                is_depot=is_depot,
            )
            demand: int = self.demands.get(vertex)
            display = self.display_data.get(vertex)
            if not demand is None:
                graph[vertex]["demand"] = demand
            if not display is None:
                graph[vertex]["display"] = display
            if self.node_coords:
                coord = self.node_coords.get(vertex)
                graph.nodes[names[vertex]]["x"] = coord[0]
                graph.nodes[names[vertex]]["y"] = coord[1]

    def get_graph(self, normalize: bool = False) -> nx.Graph:
        """Return a networkx graph instance representing the problem.

        Args:
            normalize: rename nodes to be zero-indexed
        """
        # directed graphs are fundamentally different
        graph: nx.Graph = nx.Graph() if self.is_symmetric() else nx.DiGraph()

        # set up a map from original node name to new node name
        nodes: List[Vertex] = list(self.get_nodes())
        if normalize:
            names = {n: i for i, n in enumerate(nodes)}
        else:
            names = {n: n for n in nodes}

        self.__set_node_attributes(graph, names)
        self.__set_edge_attributes(graph, names)

        # add basic graph metadata
        self.__set_graph_attributes(graph)
        return graph

    def get_total_prize(self) -> int:
        """Get the sum of prize over all vertices

        Returns:
            Total prize
        """
        # pylint: disable=no-member
        return sum([value for _, value in self.get_node_score().items()])

    def get_quota(self, alpha: int) -> int:
        """The quota is alpha percent of the total prize

        Args:
            alpha: Percent of the total prize

        Returns:
            quota
        """
        if alpha > 100:
            raise ValueError("Cannot have a percent over 100 for alpha")
        if alpha < 0:
            raise ValueError("Cannot have a negative percent for alpha")
        return int(float(alpha * self.get_total_prize()) / 100.0)

    def number_of_nodes(self) -> int:
        """Get the number of nodes in the problem

        Returns:
            Number of nodes in graph
        """
        return len(list(self.get_nodes()))

    def get_cost_limit(self) -> int:
        """Get the cost limit for a TSP with Profits problem

        Returns:
            Cost limit
        """
        return self.cost_limit

    def get_node_score(self) -> VertexLookup:
        """Get the node scores (profits)

        Returns:
            Mapping from node to node score (profit)
        """
        score_dict: VertexLookup = {}
        for key, value in self.node_score.items():  # pylint: disable=no-member
            score_dict[key] = value
        return score_dict

    def get_tsp_optimal_value(self) -> int:
        """Get the value of the optimal solution to TSP

        Returns:
            TSP optimal value
        """
        return self.tspsol

    def get_root_vertex(self, normalize: bool = False) -> Vertex:
        """Get the root vertex

        Args:
            normalize: If true, vertices start at index 0

        Returns:
            The first depot in the list

        Raises:
            ValueError: If the list of depots is empty
        """
        nodes: List[Vertex] = list(self.get_nodes())
        if normalize:
            names = {n: i for i, n in enumerate(nodes)}
        else:
            names = {n: n for n in nodes}
        try:
            # pylint: disable=unsubscriptable-object
            return names[self.depots[0]]
        except KeyError as key_error:
            raise ValueError("The list of depots is empty") from key_error

    # pylint: disable=arguments-differ
    def get_edges(self, normalize: bool = False) -> EdgeList:
        """Get a list of edges in the graph

        Args:
            normalize: If true use the normalized vertex ids

        Returns:
            List of edges in the graph
        """
        if normalize:
            raise NotImplementedError("Normalizing edges not yet implemented")
        return list(super().get_edges())


def is_pctsp_yes_instance(
    graph: nx.Graph, quota: int, root_vertex: Vertex, edge_list: EdgeList
) -> bool:
    """Returns true if the list of edges is a solution to the instance
    of the Prize collecting Travelling Salesman Problem.

    Args:
        graph: Undirected graph with cost function on edges and prize function on vertices
        quota: The salesman must collect at least the quota in prize money
        root_vertex: Start and finish vertex of the tour
        edge_list: Edges in the solution of the instance

    Returns:
        True if the total prize of the tour is at least the quota and the tour is a simple
        cycle that starts and ends at the root vertex. False otherwise.
    """
    if len(edge_list) < 3:
        return False
    walk = walk_from_edge_list(edge_list)
    vertex_set = set(walk)
    return (
        is_simple_cycle(graph, walk)
        and total_prize(
            nx.get_node_attributes(graph, VertexFunctionName.prize.value), vertex_set
        )
        >= quota
        and root_vertex == walk[0]
        and root_vertex == walk[len(walk) - 1]
    )



"""Fixes for tsplib95"""

from tsplib95 import fields, transformers


class TupleT(transformers.ContainerT):
    """Transformer for tuples"""

    def pack(self, items):
        return tuple(items)

    def unpack(self, container):
        return tuple(container)


class TempAdjacencyListField(fields.TransformerField):
    """Field for an adjacency list."""

    default = dict

    @classmethod
    def build_transformer(cls):
        """Build transformer"""
        return transformers.MapT(
            key=transformers.FuncT(func=int),
            value=transformers.ListT(value=transformers.FuncT(func=int), terminal="-1"),
            sep="\n",
            terminal="-1",
        )


class TempEdgeListField(fields.TransformerField):
    """Field for a list of edges."""

    default = list

    @classmethod
    def build_transformer(cls):
        edge = TupleT(value=transformers.FuncT(func=int), size=2)
        return transformers.ListT(value=edge, terminal="-1", sep="\n")


class TempEdgeDataField(fields.TransformerField):
    """Field for edge data."""

    default = dict

    @classmethod
    def build_transformer(cls):
        adj_list = TempAdjacencyListField.build_transformer()
        edge_list = TempEdgeListField.build_transformer()
        return transformers.UnionT(edge_list, adj_list)



"""Script for generating a tsplib style txt file from londonaq CSV"""

import itertools
import json
from pathlib import Path

import networkx as nx
import pandas as pd
import typer

from urbanroute import frames_from_urbanair_api

from tspwplib import split_graph_from_properties
from tspwplib.problem import BaseTSP
from tspwplib.types import (
    EdgeWeightFormat,
    LondonaqGraphName,
    LondonaqLocation,
    LondonaqLocationShort,
    LondonaqTimestamp,
    LondonaqTimestampId,
)
from tspwplib.utils import londonaq_comment, londonaq_graph_name

OLD_EDGE_LOOKUP_JSON = "old_edge_lookup.json"
OLD_NODE_LOOKUP_JSON = "old_node_lookup.json"


def csv_from_urbanair_api(
    username: str,
    password: str,
    location: LondonaqLocation,
    distance: int,
    timestamp: LondonaqTimestamp,
    csv_output_dir: Path,
    csv_prefix: str,
) -> None:
    """Write CSV files of nodes and edges from urbanair API."""
    if location == LondonaqLocation.id:
        root_priority = "coordinate_centrality"
    else:
        root_priority = "betweenness_centrality"
    csv_output_dir.mkdir(exist_ok=True, parents=False)
    edges_df, nodes_df, pollution_df = frames_from_urbanair_api(
        username,
        password,
        location.value,
        distance,
        timestamp.value,
        root_priority=root_priority,
    )
    # output is a CSV file representing an undirected simple graph
    edges_df.to_csv(csv_output_dir / (csv_prefix + "_edges.csv"), index=False)
    nodes_df.to_csv(
        csv_output_dir / (csv_prefix + "_nodes.csv"), index=True, index_label="node"
    )
    pollution_df.to_csv(csv_output_dir / (csv_prefix + "_hexgrid.csv"), index=False)


def generate_londonaq_dataset(
    dataset_dir: Path,
    name: LondonaqGraphName,
    comment: str,
    edges_csv_filename: str = "edges.csv",
    nodes_csv_filename: str = "nodes.csv",
    old_edge_lookup: str = OLD_EDGE_LOOKUP_JSON,
    old_node_lookup: str = OLD_NODE_LOOKUP_JSON,
) -> BaseTSP:
    """Generate a londonaq dataset"""

    # get the CSV files for edges and nodes
    dataset_dir.mkdir(parents=False, exist_ok=True)
    edges_filepath = dataset_dir / edges_csv_filename
    nodes_filepath = dataset_dir / nodes_csv_filename
    if not edges_filepath.exists():
        raise FileNotFoundError(edges_filepath)
    if not nodes_filepath.exists():
        raise FileNotFoundError(nodes_filepath)
    nodes_df = pd.read_csv(nodes_filepath)
    nodes_df = nodes_df.set_index("node")
    edges_df = pd.read_csv(edges_filepath)

    # split edges then relabel the nodes
    edges_df = edges_df.set_index(["source", "target", "key"])
    edge_attrs = edges_df.to_dict("index")
    split_graph = split_graph_from_properties(
        edge_attrs,
        edge_attr_to_split="cost",
        edge_attr_to_vertex="length",
        new_vertex_attr="demand",
        old_edge_attr="old_edge",
    )
    normalize_map = {node: i for i, node in enumerate(split_graph.nodes())}
    normalized_graph = nx.relabel_nodes(split_graph, normalize_map, copy=True)

    # save the node and edge mappings to a json file
    old_edges = {
        (normalize_map[u], normalize_map[v]): data["old_edge"]
        for u, v, data in split_graph.edges(data=True)
    }
    old_vertices = {new: old for old, new in normalize_map.items()}

    # convert tuples to lists when dumping
    json_old_edges = {}
    for (u, v), old_edge in old_edges.items():
        if u in json_old_edges:
            json_old_edges[u][v] = list(old_edge)
        else:
            json_old_edges[u] = {v: old_edge}

    with open(dataset_dir / old_edge_lookup, "w", encoding="UTF-8") as json_file:
        json.dump(json_old_edges, json_file)
    with open(dataset_dir / old_node_lookup, "w", encoding="UTF-8") as json_file:
        json.dump(old_vertices, json_file)

    # get depots
    depots = list(nodes_df.loc[nodes_df.is_depot].index.map(normalize_map))
    nx.set_node_attributes(normalized_graph, False, "is_depot")
    for v in depots:
        normalized_graph.nodes[v]["is_depot"] = True

    # round cost and demand
    for _, data in normalized_graph.nodes(data=True):
        data["demand"] = round(data["demand"])
    for _, _, data in normalized_graph.edges(data=True):
        data["cost"] = round(data["cost"])

    # NOTE (not implemented yet) get node co-ordinates

    # get TSP representation
    tsp = BaseTSP.from_networkx(
        name,
        comment,
        "PCTSP",
        normalized_graph,
        edge_weight_format=EdgeWeightFormat.FULL_MATRIX,
        weight_attr_name="cost",
    )
    # save to yaml file
    tsp.to_yaml(dataset_dir / f"{name}.yaml")
    return tsp


def to_pandas_nodelist(G: nx.Graph) -> pd.DataFrame:
    """Move node attributes to a pandas dataframe. Node ID is stored in 'node' column."""
    return pd.DataFrame([{"node": node, **data} for node, data in G.nodes(data=True)])


app = typer.Typer(name="generate")

LONDONAQ_TINY_QUOTA = 500
LONDONAQ_QUOTA = 3000


@app.command(name="onlyone")
def generate_only_one_dataset(
    username: str,
    password: str,
    location_short: LondonaqLocationShort,
    timestamp_id: LondonaqTimestampId,
    dataset_dir: Path,
):
    """Generate only one dataset given the location and timestamp"""
    timestamp = LondonaqTimestamp[timestamp_id]
    location = LondonaqLocation[location_short]
    name = londonaq_graph_name(location, timestamp)
    comment = londonaq_comment(location, timestamp)
    if location_short == LondonaqLocationShort.tiny:
        distance = LONDONAQ_TINY_QUOTA
    else:
        distance = LONDONAQ_QUOTA

    csv_from_urbanair_api(
        username,
        password,
        location,
        distance,
        timestamp,
        dataset_dir / name.value,
        name,
    )
    generate_londonaq_dataset(
        dataset_dir / name.value,
        name,
        comment,
        edges_csv_filename=name.value + "_edges.csv",
        nodes_csv_filename=name.value + "_nodes.csv",
    )


@app.command(name="all")
def generate_all_datasets(username: str, password: str, dataset_dir: Path):
    """Generate all urbanair datasets"""
    for timestamp_id, location in itertools.product(
        LondonaqTimestampId, LondonaqLocationShort
    ):
        generate_only_one_dataset(
            username, password, location, timestamp_id, dataset_dir
        )


if __name__ == "__main__":
    app()



"""Tests for splitting edges"""

from tspwplib.converter import (
    split_edges,
    split_graph_from_properties,
    lookup_from_split,
    lookup_to_split,
)


def test_split_edges():
    """Test split edges"""
    edge_list = [(0, 1), (1, 2), (0, 2)]
    splits = split_edges(edge_list)
    assert len(splits) == len(edge_list) * 2
    assert (0, -1) in splits
    assert (0, -3) in splits

    # test lookups
    from_split = lookup_from_split(edge_list, splits)
    assert from_split[(0, -1)] == (0, 1)
    assert from_split[(-1, 1)] == (0, 1)
    assert from_split[(0, -3)] == (0, 2)

    to_split = lookup_to_split(edge_list, splits)
    assert to_split[(0, 1)] == ((0, -1), (-1, 1))
    assert to_split[(1, 2)] == ((1, -2), (-2, 2))


def test_split_graph_from_properties():
    """Test split graph"""
    properties = {
        (0, 1): {"weight": 5, "cost": 3},
        (1, 2): {"weight": 1, "cost": 10},
        (0, 2): {"weight": 2, "cost": 5},
    }
    G = split_graph_from_properties(properties)
    for _, _, data in G.edges(data=True):
        old_edge = data["old_edge"]
        assert data["cost"] == float(properties[old_edge]["cost"]) or data["cost"] == 0



"""Tests for converting to an undirected graph"""

import networkx as nx
from tspwplib.converter import (
    biggest_vertex_id_from_graph,
    new_dummy_vertex,
    old_vertex_from_dummy,
    to_simple_undirected,
)


def test_new_dummy_vertex():
    """New dummy vertex test"""
    biggest = 2
    assert new_dummy_vertex(0, 0, biggest) == 0
    assert new_dummy_vertex(0, 1, biggest) == -1
    assert new_dummy_vertex(0, 2, biggest) == -4
    assert new_dummy_vertex(1, 1, biggest) == -2
    assert new_dummy_vertex(2, 2, biggest) == -6


def test_old_dummy_vertex():
    """Test getting the old vertex ID from a dummy vertex"""
    biggest = 2
    assert old_vertex_from_dummy(0, 0, biggest) == 0
    assert old_vertex_from_dummy(-1, 1, biggest) == 0
    assert old_vertex_from_dummy(-4, 2, biggest) == 0
    assert old_vertex_from_dummy(-2, 1, biggest) == 1
    assert old_vertex_from_dummy(-6, 2, biggest) == 2


def test_to_simple_undirected(undirected_complete_graph):
    """Test converting a multi graph to a simple graph (both undirected)"""
    G = nx.MultiGraph(undirected_complete_graph)
    G.add_edge(0, 2, cost=1)
    G.add_edge(1, 2, cost=1)
    simple_graph = to_simple_undirected(G)
    for v in G:
        assert simple_graph.has_node(v)
    biggest = biggest_vertex_id_from_graph(G)
    dummy0 = new_dummy_vertex(0, 1, biggest)
    dummy1 = new_dummy_vertex(1, 1, biggest)
    assert simple_graph.has_node(dummy0)
    assert simple_graph.has_node(dummy1)
    assert simple_graph.has_edge(dummy0, 2)
    assert simple_graph.has_edge(dummy1, 2)



"""Tests for walks"""

import pytest
import networkx as nx
from tspwplib import (
    EdgesNotAdjacentException,
    NotSimpleException,
    VertexFunctionName,
    edge_list_from_walk,
    is_simple_cycle,
    is_simple_path,
    is_walk,
    order_edge_list,
    remove_self_loops_from_edge_list,
    reorder_edge_list_from_root,
    total_cost_networkx,
    total_prize,
    total_prize_of_tour,
    vertex_set_from_edge_list,
    walk_from_edge_list,
)


def test_is_open_walk_is_walk(walk_graph, open_walk):
    """Test walks"""
    assert is_walk(walk_graph, open_walk)


def test_is_not_a_walk(walk_graph, not_a_walk):
    """Test invalid walks are detected"""
    assert not is_walk(walk_graph, not_a_walk)


def test_edge_list_from_walk():
    """Test edge list from any walk"""
    assert not edge_list_from_walk([])
    assert not edge_list_from_walk([0])
    assert edge_list_from_walk([0, 1, 3, 1, 2]) == [(0, 1), (1, 3), (3, 1), (1, 2)]


def test_vertex_set_from_edge_list():
    """Test a vertex set is returned from an edge list"""
    assert vertex_set_from_edge_list([]) == set()
    assert vertex_set_from_edge_list([(0, 1), (1, 2), (2, 2)]) == {0, 1, 2}


def test_order_edge_list():
    """Test edges can be ordered"""
    edges1 = [(0, 1), (1, 2), (2, 3)]
    edges2 = [(1, 0), (2, 1), (0, 2)]
    edges3 = [(1, 0), (2, 1), (0, 2), (3, 4), (4, 5), (3, 5)]
    edges4 = [(0, 1), (2, 3), (1, 2)]
    edges5 = [(0, 2), (2, 7), (4, 6), (6, 7), (0, 1), (1, 4)]
    assert order_edge_list(edges1) == edges1
    assert order_edge_list(edges2) == [(1, 0), (0, 2), (2, 1)]
    assert order_edge_list(edges4) == edges1
    assert order_edge_list(edges5)
    with pytest.raises(NotSimpleException):
        order_edge_list(edges3)


def test_reorder_edge_list_from_root():
    """Test edges are reordered starting and ending at root"""
    root = 0
    edges1 = [(2, 3), (1, 2), (0, 1)]
    edges2 = [(2, 1), (1, 0), (0, 2)]
    edges3 = []
    edges4 = [(1, 0), (2, 1), (0, 2)]
    ordered1 = reorder_edge_list_from_root(edges1, root)
    ordered2 = reorder_edge_list_from_root(edges2, root)
    ordered4 = reorder_edge_list_from_root(edges4, root)
    assert root in ordered1[0]
    assert root in ordered2[0]
    assert root in ordered2[len(ordered2) - 1]
    with pytest.raises(nx.NodeNotFound):
        reorder_edge_list_from_root(edges3, root)
    assert root in ordered4[0]
    assert root in ordered4[len(ordered4) - 1]


def test_walk_from_edge_list():
    """Test a walk is returned from an edge list"""
    assert not walk_from_edge_list([])
    assert walk_from_edge_list([(0, 1)]) == [0, 1]
    assert walk_from_edge_list([(0, 1), (2, 1)]) == [0, 1, 2]
    assert walk_from_edge_list([(1, 0), (2, 1), (0, 2)]) == [0, 1, 2, 0]
    with pytest.raises(EdgesNotAdjacentException):
        walk_from_edge_list([(0, 1), (2, 3)])
    with pytest.raises(EdgesNotAdjacentException):
        walk_from_edge_list([(0, 1), (2, 1), (0, 3)])
    with pytest.raises(EdgesNotAdjacentException):
        walk_from_edge_list([(0, 1), (2, 1), (1, 0)])


def test_is_simple_path(walk_graph, simple_path):
    """Test simple paths"""
    assert is_simple_path(walk_graph, simple_path)


def test_walk_is_not_simple_path(walk_graph, open_walk):
    """Test walks with repeated vertices are not simple paths"""
    if len(open_walk) > 1:
        assert not is_simple_path(walk_graph, open_walk)


def test_simple_cycle_is_not_simple_path(walk_graph, simple_cycle):
    """Test cycles are not simple paths"""
    if len(simple_cycle) > 1:
        assert not is_simple_path(walk_graph, simple_cycle)


def test_is_simple_cycle(walk_graph, simple_cycle):
    """Test cycles are simple"""
    assert is_simple_cycle(walk_graph, simple_cycle)


def test_path_is_not_simple_cycle(walk_graph, simple_path):
    """Test paths are not cycles"""
    if len(simple_path) > 1:
        assert not is_simple_cycle(walk_graph, simple_path)


def test_total_cost_networkx(weighted_walk_networkx_graph):
    """Test total cost"""
    assert total_cost_networkx(weighted_walk_networkx_graph, [0, 1, 2]) == 1 + 3
    assert total_cost_networkx(weighted_walk_networkx_graph, []) == 0
    assert (
        total_cost_networkx(weighted_walk_networkx_graph, [0, 1, 2, 1, 0])
        == 1 + 3 + 3 + 1
    )
    with pytest.raises(KeyError):
        total_cost_networkx(weighted_walk_networkx_graph, [0, 1, 2, 3])


def test_total_prize(weighted_walk_networkx_graph):
    """Test total prize"""
    prizes = nx.get_node_attributes(
        weighted_walk_networkx_graph, VertexFunctionName.prize
    )
    assert total_prize(prizes, [0, 1, 2]) == 3
    assert total_prize(prizes, [1, 2, 3, 1]) == 7
    assert total_prize(prizes, []) == 0


def test_total_prize_of_tour(weighted_walk_networkx_graph):
    """Test total prize of a tour"""
    prizes = nx.get_node_attributes(
        weighted_walk_networkx_graph, VertexFunctionName.prize
    )
    assert total_prize_of_tour(prizes, [0, 1, 2, 0]) == 3
    assert total_prize_of_tour(prizes, [1, 2, 3, 1]) == 6
    assert total_prize_of_tour(prizes, []) == 0


def test_remove_self_loops_from_edge_list():
    """Test all self loops are removed"""
    edge_list = [(0, 1), (1, 3), (3, 1), (1, 2)]
    assert remove_self_loops_from_edge_list(edge_list) == edge_list
    with_self_loop = edge_list + [(5, 5), (6, 6)]
    assert remove_self_loops_from_edge_list(with_self_loop) == edge_list



"""Exceptions for Travelling Salesman Problem with Profits"""

from networkx import exception as ex


class EdgeNotFoundException(ex.NetworkXException):
    """An edge could not be found in the graph"""


class EdgesNotAdjacentException(ex.AmbiguousSolution):
    """An edge list has been given non-adjacent edges which is ambiguous"""


class NotSimpleException(ex.NetworkXException):
    """A path, cycle or walk is not simple"""


class NotSimpleCycleException(NotSimpleException):
    """The walk was not a simple cycle"""


class NotSimplePathException(NotSimpleException):
    """The walk was not a simple path"""


class UnexpectedSelfLoopException(ex.NetworkXException):
    """A self loop was found where it wasn't exprected"""


class GraphStructureException(ex.NetworkXException):
    """The structure of a graph is not allowed"""


class NoTreesException(GraphStructureException):
    """The graph is not allowed to be a tree"""


class NotConnectedException(GraphStructureException):
    """The graph must be connected"""



"""Setup script for tspwplib"""

from setuptools import setup

setup(
    author="Patrick O'Hara",
    author_email="patrick.h.o-hara@warwick.ac.uk",
    url="https://github.com/PatrickOHara/tspwplib",
    description="Library of instances for TSP with Profits",
    install_requires=[
        "networkx>=2.6.0",
        "numpy>=1.21.0",
        "pandas>=1.0.0",
        "pydantic>=1.8.2",
        "tsplib95>=0.7.1",
    ],
    name="tspwplib",
    packages=["tspwplib"],
    python_requires=">=3.7",
    license="MIT",
    classifiers=[
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3 :: Only",
        "Operating System :: OS Independent",
    ],
    use_scm_version=True,
    setup_requires=["setuptools_scm"],
)



"""Tests for sparsity"""

import pytest
from tspwplib import (
    ProfitsProblem,
    build_path_to_oplib_instance,
    sparsify_by_cost,
    sparsify_uid,
)


@pytest.mark.parametrize("k", [1, 2, 5, 10])
def test_sparsify_uid(oplib_root, generation, graph_name, k):
    """Test sparsity is created uniformly and independently"""
    filepath = build_path_to_oplib_instance(oplib_root, generation, graph_name)
    complete_problem = ProfitsProblem.load(filepath)
    complete_graph = complete_problem.get_graph()
    sparse_graph = sparsify_uid(complete_graph, k)
    assert sparse_graph.number_of_nodes() == complete_graph.number_of_nodes()
    assert sparse_graph.number_of_edges() == complete_graph.number_of_nodes() * k
    for v in complete_graph:
        if complete_graph.has_edge(v, v):
            assert sparse_graph.has_edge(v, v)


@pytest.mark.parametrize("k", [1, 2, 5, 10])
def test_sparsity_by_cost(oplib_root, generation, graph_name, k):
    """Test sparsity where edge removal is weighted by cost"""
    filepath = build_path_to_oplib_instance(oplib_root, generation, graph_name)
    complete_problem = ProfitsProblem.load(filepath)
    complete_graph = complete_problem.get_graph()
    sparse_graph = sparsify_by_cost(complete_graph, k, cost_attr="cost")
    assert sparse_graph.number_of_nodes() == complete_graph.number_of_nodes()
    assert sparse_graph.number_of_edges() == complete_graph.number_of_nodes() * k
    for v in complete_graph:
        if complete_graph.has_edge(v, v):
            assert sparse_graph.has_edge(v, v)



"""Converting between different types of graphs"""


from copy import deepcopy
import math
from typing import Dict, List, Tuple, Union
import networkx as nx
import pandas as pd
from .exception import UnexpectedSelfLoopException
from .types import (
    Edge,
    EdgeFunction,
    EdgeFunctionName,
    EdgeList,
    EdgeProperties,
    MultiEdge,
    Vertex,
    VertexFunction,
    VertexFunctionName,
    VertexList,
)


def to_vertex_dataframe(graph: nx.Graph) -> pd.DataFrame:
    """Convert graph vertices to pandas dataframe

    Args:
        graph: Input graph

    Returns:
        pandas dataframe with vertex set as index
    """
    vertex_data = list(map(lambda x: dict(vertex=x[0], **x[1]), graph.nodes(data=True)))
    vertex_df = pd.DataFrame(vertex_data)
    vertex_df = vertex_df.set_index("vertex")
    return vertex_df


def asymmetric_from_directed(G: nx.DiGraph) -> nx.DiGraph:
    """Create asymmetric directed graph from directed graph

    Split every node u two nodes u1 and u2.
    We add a directed arc between u1 and u2.
    Any previous inward edge v->u is now v->u1 and any outward edge from u is now u2->v.

    Args:
        G: Directed graph

    Returns:
        Directed asymmetric graph
    """
    asymmetric_graph = nx.DiGraph()
    nodes_for_adding: List[Tuple[Vertex, Dict]] = []
    edges_for_adding: List[Tuple[Vertex, Vertex, Dict]] = []
    # deepcopy when not a view
    asymmetric_graph.graph.update(deepcopy(G.graph))

    # find the id of the biggest vertex
    biggest_vertex = biggest_vertex_id_from_graph(G)

    for vertex, data in G.nodes(data=True):
        # split the vertex into two
        head = split_head(biggest_vertex, vertex)
        tail = split_tail(biggest_vertex, vertex)

        # the data is copied to both vertices
        tail_data = data.copy()
        head_data = data.copy()

        # split the value of the prize
        prize: int = data.get(VertexFunctionName.prize, 0)
        tail_data[VertexFunctionName.prize] = tail_prize(prize)
        head_data[VertexFunctionName.prize] = head_prize(prize)

        # add vertex to asymmetric graph with new data
        nodes_for_adding.append((head, head_data))
        nodes_for_adding.append((tail, tail_data))

        # add zero-cost edge from tail to head
        edge = (tail, head, {EdgeFunctionName.cost.value: 0})
        edges_for_adding.append(edge)

    for u, v, edge_data in G.edges(data=True):
        # add edge from head of u to tail of v with data
        u_head = split_head(biggest_vertex, u)
        v_tail = split_tail(biggest_vertex, v)
        edge_uv = (u_head, v_tail, edge_data)
        edges_for_adding.append(edge_uv)

    # add nodes and edges then return graph
    asymmetric_graph.add_nodes_from(nodes_for_adding)
    asymmetric_graph.add_edges_from(edges_for_adding)
    return asymmetric_graph


def asymmetric_from_undirected(G: nx.Graph) -> nx.DiGraph:
    """Create asymmetric directed graph from undirected graph

    Args:
        G: Undirected graph

    Returns:
        Directed asymmetric graph
    """
    directed_graph = G.to_directed()
    return asymmetric_from_directed(directed_graph)


def biggest_vertex_id_from_graph(G: nx.Graph) -> Vertex:
    """Return the vertex with the largest integer id

    Args:
        G: Graph

    Returns:
        Vertex with biggest id
    """
    return max(G)


def get_original_from_split_vertex(
    biggest_vertex: Vertex, split_vertex: Vertex
) -> Vertex:
    """Return the original vertex id given a split vertex (may be head or tail)

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        split_vertex: A split vertex in asymmetric graph

    Returns:
        ID of the vertex in the original graph
    """
    if is_vertex_split_head(biggest_vertex, split_vertex):
        return split_vertex - 2 * (biggest_vertex + 1)
    # else split tail
    return split_vertex - biggest_vertex - 1


def get_original_path_from_split_path(
    biggest_vertex: Vertex, split_path: VertexList
) -> VertexList:
    """Get the path in the original graph given a path of split vertices in the asymmetric graph

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        split_path: A path of split vertices in the asymmetric directed graph

    Returns:
        A path of vertices in the original graph
    """
    original_path = []
    previous_vertex = -1
    for split_vertex in split_path:
        original_vertex = get_original_from_split_vertex(biggest_vertex, split_vertex)
        if is_vertex_split_tail(biggest_vertex, split_vertex):
            original_path.append(original_vertex)
        elif (
            is_vertex_split_head(biggest_vertex, split_vertex)
            and previous_vertex != original_vertex
        ):
            original_path.append(original_vertex)
        previous_vertex = original_vertex
    return original_path


def is_split_vertex_pair(biggest_vertex: Vertex, tail: Vertex, head: Vertex) -> bool:
    """Does the arc (tail, head) represent a split vertex in the original graph?

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        tail: Tail of edge in directed graph
        head: Head of edge in directed graph

    Returns:
        True if the arc (tail, head) represents a split vertex in the original graph
    """
    return (
        head - tail == biggest_vertex + 1
        and is_vertex_split_head(biggest_vertex, head)
        and is_vertex_split_tail(biggest_vertex, tail)
    )


def is_vertex_split_tail(biggest_vertex: Vertex, vertex: Vertex) -> bool:
    """Is the vertex a tail in the asymmetric graph?

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        vertex: A potential tail of an edge in directed graph

    Returns:
        True if the vertex is a tail
    """
    return biggest_vertex + 1 <= vertex < 2 * (biggest_vertex + 1)


def is_vertex_split_head(biggest_vertex: Vertex, split_vertex: Vertex) -> bool:
    """Is the vertex a head in the asymmetric graph?

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        split_vertex: A potential head of an edge in directed graph

    Returns:
        True if the vertex is a head
    """
    return 2 * (biggest_vertex + 1) <= split_vertex < 3 * (biggest_vertex + 1)


def split_head(biggest_vertex: Vertex, original_vertex: Vertex) -> Vertex:
    """Get the split head of the vertex

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        original_vertex: Vertex in the original graph

    Returns:
        New split vertex that is a head of all arcs in the asymmetric graph
    """
    return 2 * (biggest_vertex + 1) + original_vertex


def split_tail(biggest_vertex: Vertex, original_vertex: Vertex) -> Vertex:
    """Get the split tail of the vertex

    Args:
        biggest_vertex: The vertex with the biggest id in the original graph
        original_vertex: Vertex in the original graph

    Returns:
        New split vertex that is a tail of all arcs in the asymmetric graph
    """
    return biggest_vertex + 1 + original_vertex


def head_prize(prize: int) -> int:
    """Get the prize of the split head

    Args:
        prize: The prize of a vertex

    Returns
        Split head prize
    """
    if prize % 2 == 1:
        return math.ceil(prize / 2.0)
    return int(prize / 2)


def tail_prize(prize: int) -> int:
    """Get the prize of the split tail

    Args:
        prize: The prize of a vertex

    Returns
        Split tail prize
    """
    return math.floor(prize / 2.0)


def new_dummy_vertex(vertex: int, key: int, biggest: int) -> int:
    """New dummy vertex ID

    Args:
        vertex: Vertex ID
        key: Edge key
        biggest: Biggest vertex ID

    Returns:
        ID of a new negative dummy vertex if key is greater than one.
        Otherwise return the same vertex ID as the input.
    """
    if key > 0:
        return -((biggest + 1) * (key - 1)) - vertex - 1
    return vertex


def old_vertex_from_dummy(dummy: int, key: int, biggest) -> int:
    """Old vertex ID from the dummy vertex ID"""
    if dummy < 0:
        return -(dummy + (biggest + 1) * (key - 1) + 1)
    return dummy


def to_simple_undirected(G: nx.MultiGraph) -> nx.Graph:
    """Given an undirected multigraph, multi edges to create a simple undirected graph.

    Args:
        G: Undirected networkx multi graph.

    Returns:
        Undirected networkx simple graph with no multi edges.

    Notes:
        Assumes the vertex ids are integers.
    """
    if not isinstance(G, nx.MultiGraph) and isinstance(G, nx.Graph):
        return G
    if isinstance(G, nx.DiGraph):
        raise TypeError("Directed graphs are not valid for this method")
    simple_graph = nx.Graph()

    # copy graph attributes to new graph
    for key, value in simple_graph.graph.items():
        simple_graph.graph[key] = value

    # copy vertex attributes
    for v, data in G.nodes(data=True):
        simple_graph.add_node(v, **data)

    biggest = biggest_vertex_id_from_graph(G)

    for u, v, k, data in G.edges.data(keys=True):
        if u == v and k > 0:
            message = "Self loop found with key greater than zero: "
            message += "implies there is more than one self loop on this vertex."
            raise UnexpectedSelfLoopException(message)

        # the first multi edge - add all data to new graph edge
        if k == 0:
            simple_graph.add_edge(u, v, **data)

        # multi edge - create new vertex for the source if it does not yet exist
        elif k > 0:
            vertex_data = G.nodes[u]
            dummy = new_dummy_vertex(u, k, biggest)
            simple_graph.add_node(dummy, **vertex_data)
            simple_graph.add_edge(u, dummy, **data)
            simple_graph.add_edge(dummy, v, **data)
        else:
            raise ValueError("Negative key for edge.")

    return simple_graph


def split_edges(edge_list: EdgeList) -> List[Edge]:
    """Split each edge (u,v) by adding a new vertex w and two new edges (u,w), (w,v).

    Args:
        edge_list: List of edges or multi-edges

    Returns:
        List of edges (size 2 tuple).
        Size of returned edge list is twice the size of the input edges.
    """
    new_vertex = -1
    splits: List[Edge] = []
    for edge in edge_list:
        splits.append((edge[0], new_vertex))
        splits.append((new_vertex, edge[1]))
        new_vertex -= 1
    return splits


LookupFromSplit = Dict[Edge, Union[Edge, MultiEdge]]


def lookup_from_split(edge_list: EdgeList, splits: List[Edge]) -> LookupFromSplit:
    """Get lookup from a split edge to an original edge.

    Args:
        edge_list: Edge in original graph.
        splits: List of edges created by [split_edges][tspwplib.converter.split_edges].

    Returns:
        Dictionary lookup from split edges to the original edges.
    """
    lookup = {}
    for i, edge in enumerate(edge_list):
        lookup[splits[2 * i]] = edge
        lookup[splits[2 * i + 1]] = edge
    return lookup


LookupToSplit = Dict[Union[Edge, MultiEdge], Tuple[Edge, Edge]]


def lookup_to_split(edge_list: EdgeList, splits: List[Edge]) -> LookupToSplit:
    """Get lookup from an original edge to the two split edges.

    Args:
        edge_list: Edge in original graph.
        splits: List of edges created by [split_edges][tspwplib.converter.split_edges].

    Returns:
        Dictionary lookup from the original edges to a pair of split edges.
    """
    lookup = {}
    for i, edge in enumerate(edge_list):
        lookup[edge] = (splits[2 * i], splits[2 * i + 1])
    return lookup


def prize_from_weighted_edges(
    edge_weights: EdgeFunction, to_split: LookupToSplit
) -> VertexFunction:
    """Get a prize function on the vertices from a weight function on the edges.

    Args:
        edge_weights: Lookup from edges to weights.
        to_split: Lookup from original edges to pairs of split edges
            (see [lookup_to_split][tspwplib.converter.lookup_to_split]).

    Returns:
        Lookup from fake vertices to weight of original edge that the fake vertex represents.
    """
    prizes = {}
    for edge, weight in edge_weights.items():
        first_split, second_split = to_split[edge]
        if first_split[1] != second_split[0]:
            message = "Second vertex of first edge and first vertex of second edge "
            message += "must match in to_split_lookup"
            raise LookupError(message)
        vertex = first_split[1]
        prizes[vertex] = weight
    return prizes


def split_edge_cost(
    edge_cost: EdgeFunction, to_split: LookupToSplit
) -> Dict[Edge, int]:
    """Assign half the cost of the original edge to each of the split edges.

    Args:
        edge_cost: Lookup from edges to cost.
        to_split: Lookup from original edges to pairs of split edges
            (see [lookup_to_split][tspwplib.converter.lookup_to_split]).

    Returns:
        Lookup from split edges to cost.

    Notes:
        The cost is cast to a float.
    """
    split_cost = {}
    for edge, cost in edge_cost.items():
        first_split, second_split = to_split[edge]
        split_cost[first_split] = cost
        split_cost[second_split] = 0
    return split_cost


def split_graph_from_properties(
    edge_properties: EdgeProperties,
    edge_attr_to_split: str = "cost",
    edge_attr_to_vertex: str = "length",
    new_vertex_attr: str = "prize",
    old_edge_attr: str = "old_edge",
) -> nx.Graph:
    """Split edges with properties and create undirected simple graph.

    Args:
        edge_properties: Keys are edges. Values are dicts of edge attributes.
        edge_attr_to_split: Name of edge attribute. Assign half the value to each split edge.
        edge_attr_to_vertex: Name of edge attribute. Assign edge value to a new vertex attribute.
        new_vertex_attr: Name of the newly created vertex attribute.
        old_edge_attr: Name of the newly created attribute for the old edge ID.

    Returns:
        Undirected simple graph with edge attributes for cost, prize and old_edge

    Notes:
        To get the original_edge that a split edge represents, access the 'old_edge' attribute
    """
    # check that every edge has an attribute to split and an attr to move to vertex
    is_edge_attr_to_split = True
    is_edge_attr_to_vertex = True
    for data in edge_properties.values():
        if not edge_attr_to_split in data:
            is_edge_attr_to_split = False
        if not edge_attr_to_vertex in data:
            is_edge_attr_to_vertex = False

    # split edges and create lookups
    edge_list = list(edge_properties.keys())
    splits = split_edges(edge_list)
    to_split = lookup_to_split(edge_list, splits)
    from_split = lookup_from_split(edge_list, splits)

    # create graph and assign prizes and costs
    G = nx.Graph()
    G.add_edges_from(splits)
    if is_edge_attr_to_vertex:
        prize = prize_from_weighted_edges(
            {edge: item[edge_attr_to_vertex] for edge, item in edge_properties.items()},
            to_split,
        )
        nx.set_node_attributes(G, 0.0, name=new_vertex_attr)
        nx.set_node_attributes(G, prize, name=new_vertex_attr)

    if is_edge_attr_to_split:
        cost = split_edge_cost(
            {edge: item[edge_attr_to_split] for edge, item in edge_properties.items()},
            to_split,
        )
        nx.set_edge_attributes(G, 0.0, name=edge_attr_to_split)
        nx.set_edge_attributes(G, cost, name=edge_attr_to_split)
    nx.set_edge_attributes(G, from_split, name=old_edge_attr)
    return G



"""Functions for creating a sparse graph from a complete graph"""

from copy import deepcopy
import random
import networkx as nx
from .types import Vertex


def sparsify_uid(
    G: nx.Graph,
    kappa: int,
    remove_self_loops: bool = False,
    seed: int = 0,
) -> nx.Graph:
    """Remove edges with uniform and independent (uid) probability until
    the number of edges equals kappa * number of nodes

    Args:
        G: Graph
        kappa: Parameter independent of the input size
        remove_self_loops: Should self loops have a change of being removed?
        seed: Set the random seed

    Returns:
        Graph with kappa * V edges where V is the number of nodes

    Notes:
        A copy of the graph is made and returned. The original graph is unedited.
    """
    random.seed(seed)
    graph_copy = deepcopy(G)
    vertex_list = list(graph_copy.nodes())
    while graph_copy.number_of_edges() > graph_copy.number_of_nodes() * kappa:
        # choose vertex randomly
        u = random.choice(vertex_list)
        if graph_copy.degree(u) > 0:
            # choose a neighbor randomly
            v = random.choice(list(graph_copy.neighbors(u)))
            # remove edge if it is not a self loop
            if remove_self_loops or u != v:
                graph_copy.remove_edge(u, v)
    return graph_copy


def total_cost_of_adjacent_edges(
    G: nx.Graph, u: Vertex, cost_attr: str = "cost"
) -> int:
    """Get the total cost of edges adjacent to a vertex u"""
    total = 0
    for v in G.neighbors(u):
        total += G.get_edge_data(u, v)[cost_attr]
    return total


def sparsify_by_cost(
    G: nx.Graph,
    kappa: int,
    cost_attr: str = "cost",
    seed: int = 0,
    remove_self_loops: bool = False,
) -> nx.Graph:
    """Given vertex i, remove an edge e=(i,j) with probability P[i,j]
    where the probability function is weighted according to the cost function:

    $P[i,j] = c(i,j) / C_i$

    where $C_i$ is the total cost of all edges adjacent to vertex i.

    Args:
        G: Graph
        kappa: Parameter independent of the input size
        cost_attr: Name of the cost attribute on edges
        remove_self_loops: Should self loops have a change of being removed?
        seed: Set the random seed for reproducibility of graphs

    Returns:
        A deep copy of the original graph with edges removed
    """
    random.seed(seed)
    graph_copy = deepcopy(G)
    vertex_list = list(graph_copy.nodes())
    while graph_copy.number_of_edges() > graph_copy.number_of_nodes() * kappa:
        # choose vertex randomly
        u = random.choice(vertex_list)
        u_has_self_loop = graph_copy.has_edge(u, u)
        u_neighbors = list(graph_copy.neighbors(u))
        if u_has_self_loop and not remove_self_loops:
            u_neighbors.remove(u)
        if len(u_neighbors) > 0:
            # draw a random number from uniform and independent distribution (UID)
            threshold = random.random()

            # the total cost of all adjacent edges is the denominator
            cost_of_edges = float(
                total_cost_of_adjacent_edges(graph_copy, u, cost_attr=cost_attr)
            )
            chosen_vertex = None

            if u_has_self_loop and not remove_self_loops:
                cost_of_edges -= graph_copy.get_edge_data(u, u)[cost_attr]

            if cost_of_edges == 0:
                chosen_vertex = random.choice(u_neighbors)

            else:
                # iterate over neighbors to check if the edge has been selected
                probability_so_far = 0.0
                for v in u_neighbors:
                    # weight probability of choosing an edges by the cost
                    edge_prob = (
                        float(graph_copy.get_edge_data(u, v)[cost_attr]) / cost_of_edges
                    )
                    if probability_so_far + edge_prob >= threshold and (
                        remove_self_loops or u != v
                    ):
                        # edge found! break here
                        chosen_vertex = v
                        break
                    probability_so_far += edge_prob
            graph_copy.remove_edge(u, chosen_vertex)
    return graph_copy



"""Tests for the pydantic representation of a TSP"""

import pytest
from tsplib95.models import StandardProblem
from tspwplib import BaseTSP, EdgeWeightType, GraphName, build_path_to_tsplib_instance


def test_get_weighted_full_matrix(tsplib_root, graph_name):
    """Test weighted adjacency matrix"""
    tsp_path = build_path_to_tsplib_instance(tsplib_root, graph_name)
    problem = StandardProblem.load(tsp_path)
    tsp = BaseTSP.from_tsplib95(problem)
    weights = tsp.get_weighted_full_matrix()
    for i, u in enumerate(problem.get_nodes()):
        for j, v in enumerate(problem.get_nodes()):
            assert weights[i, j] == problem.get_weight(u, v)


@pytest.mark.parametrize("gname", list(GraphName))
def test_from_tsplib95(tsplib_root, gname):
    """Test tsplib95 problems can be read into BaseTSP"""
    # only load problems with less than 1000 vertices
    n_nodes = int("".join(filter(str.isdigit, gname.value)))
    if n_nodes < 1000:
        tsp_path = build_path_to_tsplib_instance(tsplib_root, gname)
        assert tsp_path.exists()
        problem = StandardProblem.load(tsp_path)
        tsp = BaseTSP.from_tsplib95(problem)
        assert len(tsp.edge_data) == len(list(problem.get_edges()))
        for edge, weight in tsp.edge_weights.items():
            assert weight == problem.get_weight(edge[0], edge[1])


@pytest.mark.parametrize("gname", list(GraphName))
def test_to_tsplib95(tsplib_root, gname):
    """Test going to tsplib"""
    # only load problems with less than 1000 vertices
    n_nodes = int("".join(filter(str.isdigit, gname.value)))
    if n_nodes < 1000:
        tsp_path = build_path_to_tsplib_instance(tsplib_root, gname)
        assert tsp_path.exists()
        og_problem = StandardProblem.load(tsp_path)
        tsp = BaseTSP.from_tsplib95(og_problem)
        new_problem = tsp.to_tsplib95()
        # pylint: disable=not-an-iterable,unsupported-membership-test
        for u, v in og_problem.edge_data:
            assert (u, v) in new_problem.edge_data
            assert og_problem.get_weight(u, v) == new_problem.get_weight(u, v)


def test_uniform_random_cost(tsplib_root, graph_name):
    """Test creating a cost function with uniform random costs"""
    tsp_path = build_path_to_tsplib_instance(tsplib_root, graph_name)
    problem = StandardProblem.load(tsp_path)
    problem.edge_weight_type = EdgeWeightType.UNIFORM_RANDOM
    tsp = BaseTSP.from_tsplib95(problem)
    match_count = 0
    for (u, v), cost in tsp.edge_weights.items():
        assert 1 <= cost <= 100
        if u != v:
            if cost == problem.get_weight(u, v):
                match_count += 1
    assert match_count < 0.5 * float(len(tsp.edge_data))  # expect not many matches
